unit Cgate_classes;

interface

uses
  WinApi.Windows, System.Classes, System.SysUtils, Winapi.Messages, cgate_types,
  cgate_functions, cgate;


type
  {Mutex}
  TMutex = class(TObject)
    private
      FSection: TRTLCriticalSection;
    public
      constructor Create();
      destructor Destroy(); override;
      procedure Lock();
      procedure Unlock();
  end;

  { Cgate core Storage}
  TStStreams = array of TStorageStream;
//--- Storage
  TCGCoreStorage = class(TObject)
    private
      FStrms: TStStreams;               // Storage streams with tables
      FSymbInfo: TSymbolsTradeInfo;     // Session and instruments info
    protected
      procedure DecodeScheme(Scheme: pcg_scheme_desc; CurLstnr: TListener);
      procedure DecodeMessages(const Scheme: pcg_scheme_desc; const str_idx: integer);
      procedure FillStrData(Msg: pcg_msg; CurListn: TListener);
      procedure DeleteTables(aListener: TListener);
      procedure ClearTables(Data: pcg_data_cleardeleted; CListnr: TListener);
      procedure CopyTable(const source: TTable; var dest: TTable);
      function BuffToInt64(const buff: array of byte): int64_t;
      function BuffToString(const buff: array of byte): string;
      function BuffToInt8(const buff: array of byte): int8_t;
      procedure BuildSession(const a_idx: integer; const b_idx: integer);
      procedure UpdateSymbols();
      function BuffToTime(const buff: array of byte): tcg_time;
    public
      constructor Create();
      destructor Destroy(); override;
    property Strms: TStStreams read FStrms;
  end;

  {Connection thread}
  TCgConnThread = class(TThread)
    private
      FCSet: TP2Set;                    // Ggate settings
      FStrDone: boolean;                // Stream done
      FDoExit: boolean;                 // Do exit from thread
      FDone: boolean;                   // Thread done
      FhConn: THandle;                  // Connection handle
      FConState: uint32_t;              // Connection state
      FReplLs: TReplLisrteners;         // Repl listeners
    protected
      procedure Execute(); override;
      function MakeCallBack(const Param: ansistring; var aListener: THandle): CG_RESULT;
      function MsgCallBack(hCon: THandle; hListn: THandle; msg: pcg_msg; Data: Pointer): CG_RESULT; stdcall;
    public
      constructor Create(Sett: TP2Set);
      destructor Destroy(); override;
    property hConn: THandle read FhConn;
  end;

  {Main class Cgate core}
  TCgate_core = class(TObject)
    private
      FMutex: TMutex;                   // Mutex
      FCgDllFPath: string;              // cgate dll full path
      FCgateHandle: THandle;            // cgate dll handle
      FCgEnvir: CG_RESULT;              // Enviroiment status
      FCThread: TCgConnThread;          // connection thread
      FStorage: TCGCoreStorage;         // cgate core storage
      FProcInfo: TProcessInformation;   // router process info
    protected
      function LoadRouter(const FileName: String): Boolean;
    public
      constructor Create();
      destructor Destroy(); override;
      function Connect(CGSet: TP2Set): PAnsichar;
      function Disconnect(): boolean;
      function GetSymbols(var Data: Pointer): integer;
    property Mutex: TMutex read FMutex;
    property CgDllFPath: string read FCgDllFPath write FCgDllFPath;
  end;

var
  CgCore: TCgate_core;

implementation

//--- mutex
constructor TMutex.Create;
begin
  InitializeCriticalSection(FSection);
  inherited Create();
end;

destructor TMutex.Destroy;
begin
  DeleteCriticalSection(FSection);
  inherited Destroy;
end;

procedure TMutex.Lock;
begin
  EnterCriticalSection(FSection);
end;

procedure TMutex.Unlock;
begin
  LeaveCriticalSection(FSection);
end;
//--- end of mutex

//--- Cgate core
constructor TCgate_core.Create;
begin
  inherited Create;
  FProcInfo.hProcess:= 0;
  FProcInfo.hThread:= 0;
  FCgEnvir:= CG_ERR_NULL;
  FMutex:= TMutex.Create;
end;

destructor TCgate_core.Destroy;
begin
//--- Destroy connection, if exist and close enviroiment, if open
  if (Assigned(FCThread)) then Disconnect();
//--- mutex
  if Assigned(FMutex) then FreeAndNil(FMutex);
//--- instruments
  inherited Destroy;
end;

function TCgate_core.LoadRouter;
var
  StartInfo: TStartupInfo;
  CmdLine: String;
  RPath: string;
  k: integer;
begin
  RPath:= ExtractFilePath(Filename);
  k:= Pos('bin\', RPath);
  if (k > 0) then
  begin
   Delete(RPath, k, Length('bin\'));
   RPath:= '"' + RPath + 'CLIENT_router.ini"';
  end;
  CmdLine:= '"' + Filename + '" ' + '/ini:' + RPath;
  FillChar(StartInfo, SizeOf(StartInfo), 0);
  with StartInfo do
  begin
    cb:= SizeOf(StartInfo);
    wShowWindow := SW_SHOWNORMAL;
  end;
  Result:= CreateProcess(Nil, PWideChar(CmdLine), Nil, Nil, false,
                         NORMAL_PRIORITY_CLASS{ or CREATE_NO_WINDOW}, Nil, // TODO: comment DEBUG!!!
                         PWideChar(ExtractFilePath(Filename)), StartInfo, FProcInfo);
end;

function TCgate_core.Connect;
var
  env_string: PAnsiChar;
  major, minor, patch: integer;
  Router: string;
begin
  result:= nil;
//--- Check connection thread
  if(not Assigned(FCThread)) then
  begin
//--- Get cgate dll full path
    FCgDllFPath:= GetCgDllFullPath(string(CGSet.lpRegKey),
                                   string(CGSet.lpRegValue));
    if (CgDllFPath <> '') then
    begin
//--- Check CGATE servide running
      if (not GetCgService(CGSet.lpMachine, CGSet.lpServName)) then
      begin
//--- try load router
        Router:= ExtractFilePath(CgDllFPath);
{$ifdef CPUX64}
        Router:= Router + 'P2MQRouter64.exe';
        CgDllFPath:= CgDllFPath + 'cgate64.dll';
{$else}
        Router:= Router + 'P2MQRouter.exe';
        CgDllFPath:= CgDllFPath + 'cgate.dll';
{$endif}
        if (not LoadRouter(Router)) then
        begin
          Result:= PAnsiChar(RawByteString('Error: Spectra Gate service not running and router not found!'));
          exit;
        end;
      end else
      begin
{$ifdef CPUX64}
        CgDllFPath:= CgDllFPath + 'cgate64.dll';
{$else}
        CgDllFPath:= CgDllFPath + 'cgate.dll';
{$endif}
      end;
//--- Load cgate library
      if (LoadCgLib(CgDllFPath, FCgateHandle)) then
      begin
        if(FCgateHandle <> 0) then
        begin
//--- open envoronment
          env_string:= PansiChar(RawByteString(CGSet.lpSet.EnvStr));
          FCgEnvir:= cg_env_open(env_string);
          if(FCgEnvir = CG_ERR_OK) then
          begin
//--- Get cgate library version
            if (cg_env_getcomp_ver(PAnsiChar('cgate'), major, minor, patch) = CG_ERR_OK) then
            begin
              if (((major shl 24) + (minor shl 16) + patch) >= CG_VERSION) then
              begin
                Result:= PAnsiChar(RawByteString('Version OK!'));
                if (not Assigned(FStorage)) then FStorage:= TCGCoreStorage.Create();
                if (not Assigned(FCThread)) then FCThread:= TCgConnThread.Create(CGSet);
              end else
              begin
                if (FCgEnvir = CG_ERR_OK) then
                begin
                  if (cg_env_close() = CG_ERR_OK) then
                  begin
                    FCgEnvir:= CG_ERR_NULL;
                    if (FCgateHandle <> 0) then
                      if (FreeLibrary(FCgateHandle) = true) then
                         FCgateHandle:= 0;
                  end;
                end;
                Result:= PAnsiChar(RawByteString('Error: Cgate version must be 5.3.5 or higher!'));
              end;
            end else Result:= PAnsiChar(RawByteString('Error: Get cgate version failed!'));
          end else Result:= PAnsiChar(RawByteString('Error: Cgate enviroiment not open!'));
        end else Result:= PAnsiChar(RawByteString('Error: Cgate library not loaded!'));
      end else
      begin
        if (FCgateHandle <> 0) then
          if (FreeLibrary(FCgateHandle) = true) then FCgateHandle:= 0;
        Result:= PAnsiChar(RawByteString('Error: Cgate library not loaded!'));
      end;
    end else
      Result:= PAnsiChar(RawByteString('Error: Get cgate library path failed!'));
  end else Result:= PAnsiChar(RawByteString('Version OK!'));
end;

function TCgate_core.Disconnect;
begin
  result:= false;
//--- Destroy connection thread
  if (Assigned(FCThread)) then
  begin
    FCThread.FDoExit:= true;
    FCThread.WaitFor;
    if (FCThread.FDone = true) then
      FreeAndNil(FCThread) else exit;
  end;
//--- close envoronment and free cgate library
  if (not Assigned(FCThread)) then
  begin
    if (FCgEnvir = CG_ERR_OK) then
    begin
     if (cg_env_close() = CG_ERR_OK) then
      begin
        FCgEnvir:= CG_ERR_NULL;
        if (FCgateHandle <> 0) then
        begin
         if (FreeLibrary(FCgateHandle) = true) then
         begin
           FCgateHandle:= 0;
           Result:= true;
         end;
        end;
      end;
    end else Result:= true;
  end;
//--- Free Storage
  if (Assigned(FStorage)) then FreeAndNil(FStorage);

//--- Close router, if opened (work without service)
  if (FProcInfo.hProcess <> 0) then
  begin
    if (TerminateProcess(FProcInfo.hProcess, NO_ERROR) = true) then
    begin
      if (CloseHandle(FProcInfo.hThread) = true) then
      begin
        FProcInfo.hThread:= 0;
        if (CloseHandle(FProcInfo.hProcess) = true) then
          FProcInfo.hProcess:= 0;
      end;
    end;
  end;
end;

function TCgate_core.GetSymbols;
begin
  result:= RES_ERROR;
  CgCore.Mutex.Lock;
  try
    if ((FStorage.FSymbInfo.CurSess.sess_id > -1) and
        (FStorage.FSymbInfo.CurSess.opt_sess_id > -1)) then
    begin
      PSymbolsTradeInfo(Data)^.CurSess:= FStorage.FSymbInfo.CurSess;
//--- Symbols
      if (length(FStorage.FSymbInfo.Fut.smbls) > 0) then
      begin
        SetLength(PSymbolsTradeInfo(Data)^.Fut.smbls, length(FStorage.FSymbInfo.Fut.smbls));
        PSymbolsTradeInfo(Data)^.Fut.smbls:= FStorage.FSymbInfo.Fut.smbls;
      end;
      if (length(FStorage.FSymbInfo.Opt.smbls) > 0) then
      begin
        SetLength(PSymbolsTradeInfo(Data)^.Opt.smbls, length(FStorage.FSymbInfo.Opt.smbls));
        PSymbolsTradeInfo(Data)^.Opt.smbls:= FStorage.FSymbInfo.Opt.smbls;
      end;
      Result:= RES_OK;
    end;
  finally
    CgCore.Mutex.Unlock;
  end;
end;
//--- end of cgate core

//-- connection thread
constructor TCgConnThread.Create;
var
  i: integer;
begin
  inherited Create(false);
  FreeOnTerminate:= false;
  FCSet:= Sett;
  FDoExit:= false;
  FStrDone:= false;
  FDone:= false;
  FhConn:= 0;
  if (Sett.lpSet.LstnrCnt > 0) then
  begin
    FReplLs.Cnt:= Sett.lpSet.LstnrCnt;
    SetLength(FReplLs.Lstnrs, FReplLs.Cnt);
    for i:= 0 to FReplLs.Cnt - 1 do
    begin
      FReplLs.Lstnrs[i].Name:= GetListenerName(Sett.lpSet.ListnrsNew[i]);
      FReplLs.Lstnrs[i].OnLine:= false;
      FReplLs.Lstnrs[i].CanClose:= false;
      FReplLs.Lstnrs[i].hLstnr:= 0;
      FReplLs.Lstnrs[i].LifeNum:= 0;
      FReplLs.Lstnrs[i].OpenParam:= Sett.lpSet.LOpenParam[i];
    end;
  end;
end;

destructor TCgConnThread.Destroy;
begin
  SetLength(FReplLs.Lstnrs, 0);
  inherited Destroy;
end;

//--- Call back
function P2Lsn_CallBack(hCon: THandle; hList: THandle; msg: pcg_msg; Data: Pointer): CG_RESULT; stdcall;
begin
  Result:= TCgConnThread(Data).MsgCallBack(hCon, hList, msg, Data);
end;

function TCgConnThread.MakeCallBack;
begin
  Result:= cg_lsn_new(hConn, PAnsiChar(Param), @P2Lsn_CallBack, nil, aListener);
end;

procedure TCgConnThread.Execute();
var
  i: integer;
  h_res: CG_RESULT;
  can_close: boolean;
begin
  while (not FDoExit) do
  begin
    if (FhConn = 0) then
    begin
      h_res:= cg_conn_new(PansiChar(RawByteString(FCSet.lpSet.ConnStr)), FhConn);
      if(h_res <> CG_ERR_OK) then
        cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Error: connection not created: %d',
                                                                    [h_res]))));
    end else
    begin
      if((cg_conn_getstate(FhConn, FConState)) = CG_ERR_OK) then
      begin
        case FConState of
          CG_STATE_ERROR:
            begin
            h_res:= cg_conn_close(FhConn);
              if(h_res <> CG_ERR_OK) then
                cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Error: connection not closed: %d',
                                                                    [h_res]))));
            end;
          CG_STATE_OPENING: ; //TODO: ????
          CG_STATE_CLOSED:
            begin
              h_res:= cg_conn_open(FhConn, nil);
              if(h_res <> CG_ERR_OK) then
                cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Error: connection not open: %d',
                                                                    [h_res]))));
              if (h_res = CG_ERR_UNSUPPORTED) then
              begin
                FDoExit:= true;
                cg_log_infostr(PAnsiChar(AnsiString('--->>> Error: connection unsupported! ')));
              end;
            end;
          CG_STATE_ACTIVE:
            begin
              h_res:= cg_conn_process(FhConn, 0, nil);
              if((h_res <> CG_ERR_OK) and (h_res <> CG_ERR_TIMEOUT)) then
               cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Warning: connection state request: %d',
                                                                    [h_res]))));
//--- Repl listeners
              for i := 0 to FReplLs.Cnt - 1 do
              begin
                if (FReplLs.Lstnrs[i].hLstnr = 0) then
                begin
                  if(MakeCallBack(ansiString(FCSet.lpSet.ListnrsNew[i]),
                           FReplLs.Lstnrs[i].hLstnr) <> CG_ERR_OK) then
                  cg_log_infostr(PAnsiChar(AnsiString('--->>> Error: Listener not set callback! ')));
                end else
                begin
                  if(cg_lsn_getstate(FReplLs.Lstnrs[i].hLstnr,
                             FReplLs.Lstnrs[i].State) = CG_ERR_OK) then
                  begin
                    case FReplLs.Lstnrs[i].State of
                      CG_STATE_ERROR:
                        begin
                          cg_log_infostr(PAnsiChar(AnsiString('--->>> Error: Listener state error! ')));
                          h_res:= cg_lsn_close(FReplLs.Lstnrs[i].hLstnr);
                          if(h_res <> CG_ERR_OK) then
                             cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Error: listener not closed: %d',
                                                                    [h_res])))) else
                          begin
                            h_res:= cg_lsn_destroy(FReplLs.Lstnrs[i].hLstnr);
                            if(h_res = CG_ERR_OK) then FReplLs.Lstnrs[i].hLstnr:= 0;
                          end;
                        end;
                      CG_STATE_OPENING: ;
                      CG_STATE_CLOSED:
                        begin
                          h_res:= cg_lsn_open(FReplLs.Lstnrs[i].hLstnr,
                            PAnsiChar(RawByteString(FReplLs.Lstnrs[i].OpenParam)));
                          if(h_res <> CG_ERR_OK) then
                            cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Error: listener not open: %d',
                                                                    [h_res]))));
                        end;
                      CG_STATE_ACTIVE:;
                    end;
                  end;
                end;
              end;
            end;
        end;
      end;
    end;
  end;
//--- Close listeners and connection
  while (not FDone) do
  begin
    if (FhConn <> 0) then
    begin
      if((cg_conn_getstate(FhConn, FConState)) = CG_ERR_OK) then
      begin
        case FConState of
          CG_STATE_OPENING,
          CG_STATE_ERROR,
          CG_STATE_ACTIVE:
            begin
//--- close repl listeners
              for i := 0 to FReplLs.Cnt - 1 do
              begin
                if (FReplLs.Lstnrs[i].hLstnr <> 0) then
                begin
                  if (cg_lsn_getstate(FReplLs.Lstnrs[i].hLstnr,
                                      FReplLs.Lstnrs[i].State) = CG_ERR_OK) then
                  begin
                    case FReplLs.Lstnrs[i].State of
                      CG_STATE_ERROR:
                        begin
                          if (cg_lsn_close(FReplLs.Lstnrs[i].hLstnr) = CG_ERR_OK) then
                            if (cg_lsn_destroy(FReplLs.Lstnrs[i].hLstnr) = CG_ERR_OK) then
                              FReplLs.Lstnrs[i].hLstnr:= 0;
                        end;
                      CG_STATE_OPENING:;
                      CG_STATE_ACTIVE:
// lsn_close  conn_close      CG_MSG_OPEN. //SDK
                        if (FReplLs.Lstnrs[i].CanClose = true) then
                          cg_lsn_close(FReplLs.Lstnrs[i].hLstnr);
                      CG_STATE_CLOSED:
                        if (FReplLs.Lstnrs[i].StrDone = true) then
                          if (cg_lsn_destroy(FReplLs.Lstnrs[i].hLstnr) = CG_ERR_OK) then
                            FReplLs.Lstnrs[i].hLstnr:= 0;
                    end;
                  end;
                end;
              end;
//--- try to close connection
              can_close:= true;
              for i := 0 to FReplLs.Cnt - 1 do
              begin
                if (FReplLs.Lstnrs[i].hLstnr <> 0) then
                begin
                  can_close:= false;
                  break;
                end;
              end;
              if (can_close = true) then cg_conn_close(FhConn);
            end;
          CG_STATE_CLOSED:
            if (cg_conn_destroy(FhConn) = CG_ERR_OK) then
            begin
              FhConn:= 0;
              FDone:= true;
            end;
        end;
      end;
    end else FDone:= true;
  end;
end;

//--- P2 Cgate msg CallBack
function TCgConnThread.MsgCallBack;
var
  Scheme: pcg_scheme_desc;
  LifeNum: pcg_data_lifenum;
  ClrDel: pcg_data_cleardeleted;
  CurListener: TListener;
  CurThread: TCgConnThread;
  i, mem_idx: integer;
  aParam: PAnsiChar;
begin
//--- detect current listener
  CurThread:= TCgConnThread(CurrentThread);
  mem_idx:= -1;
  for i:= 0 to CurThread.FReplLs.Cnt - 1 do
  begin
    if (CurThread.FReplLs.Lstnrs[i].hLstnr = hListn) then
    begin
      CurListener:= CurThread.FReplLs.Lstnrs[i];
      mem_idx:= i;
      break;
    end;
  end;
//--- Recive messages
  case Msg^.msg_type of
    CG_MSG_OPEN:
      begin
        if (mem_idx >= 0) then
          CurThread.FReplLs.Lstnrs[mem_idx].CanClose:= true;    //SDK
        if (cg_lsn_getscheme(CurListener.hLstnr, Scheme) = CG_ERR_OK) then
          CGCore.FStorage.DecodeScheme(Scheme, CurListener);
      end;
    CG_MSG_CLOSE:
      if (mem_idx >= 0) then
      begin
        CurThread.FReplLs.Lstnrs[mem_idx].StrDone:= true;
        CurThread.FReplLs.Lstnrs[mem_idx].OnLine:= false;
      end;
    //CG_MSG_TIME:; no in SDK ???
    CG_MSG_DATA:; //    ()
    CG_MSG_STREAM_DATA:
      if(Msg^.data_size > 0) then
        CGCore.FStorage.FillStrData(Msg, CurListener);
   // CG_MSG_DATAARRAY:; no in SDK ???
   // CG_MSG_OBJFAILED:; no in SDK ???
   // CG_MSG_PARSEERR:; no in SDK ???
    CG_MSG_TN_BEGIN:;
    CG_MSG_TN_COMMIT:;
   // CG_MSG_TN_ROLLBACK:;  no in SDK ???
   // CG_MSG_P2MQ_RANGE_START:; no in SDK ???
    CG_MSG_P2MQ_TIMEOUT:;
   // CG_MSG_P2REPL_RANGE_START:; no in SDK ???
    CG_MSG_P2REPL_LIFENUM:
      begin
        LifeNum:= pcg_data_lifenum(Msg^.data);
        if(CurThread.FReplLs.Lstnrs[mem_idx].LifeNum <> LifeNum^.life_number) then
        begin
          CurThread.FReplLs.Lstnrs[mem_idx].LifeNum:= LifeNum^.life_number;
          if(LifeNum^.flags <> CG_LIFENUM_CHANGE__DONT_CLEAR_STORAGE) then
          begin
            if (mem_idx >= 0) then
              CurThread.FReplLs.Lstnrs[mem_idx].OnLine:= false;
            CgCore.FStorage.DeleteTables(CurListener);
          end;
        end;
      end;
    CG_MSG_P2REPL_CLEARDELETED:
      begin
        if (mem_idx >= 0) then
          CurThread.FReplLs.Lstnrs[mem_idx].OnLine:= false;
        ClrDel:= pcg_data_cleardeleted(Msg^.data);
        CgCore.FStorage.ClearTables(ClrDel, CurListener);
      end;
    CG_MSG_P2REPL_ONLINE:
      if (mem_idx >= 0) then
      begin
        CurThread.FReplLs.Lstnrs[mem_idx].OnLine:= true;
        if ((CurThread.FReplLs.Lstnrs[mem_idx].Name = 'FORTS_FUTINFO_REPL') or
            (CurThread.FReplLs.Lstnrs[mem_idx].Name = 'FORTS_OPTINFO_REPL')) then
        CgCore.FStorage.UpdateSymbols();
      end;
    CG_MSG_P2REPL_REPLSTATE:
      if (Msg^.data_size > 0) then
      begin
        aParam:= Msg^.data;
        if (mem_idx >= 0) then
          CurThread.FReplLs.Lstnrs[mem_idx].OpenParam:= 'replstate=' + string(aParam);
      end;
  end;
  result:= CG_ERR_OK;
end;
//-- end of connection thread

//-- cgcore storage
constructor TCGCoreStorage.Create;
begin
  inherited Create();
  SetLength(FStrms, 0);
  FSymbInfo.CurSess.sess_id:= -1;
  FSymbInfo.CurSess.opt_sess_id:= -1;
  SetLength(FSymbInfo.Fut.smbls, 0);
  SetLength(FSymbInfo.Opt.smbls, 0);
  FSymbInfo.Fut.sess_id:= -1;
  FSymbInfo.Opt.sess_id:= -1;
end;

destructor TCGCoreStorage.Destroy;
begin
  SetLength(FStrms, 0);
  SetLength(FSymbInfo.Fut.smbls, 0);
  SetLength(FSymbInfo.Opt.smbls, 0);
  inherited Destroy;
end;

procedure TCGCoreStorage.DecodeMessages;
var
  MsgDesc: tcg_message_desc;
  FieldDesc: tcg_field_desc;
  i, j: integer;
begin
  SetLength(FStrms[str_idx].StrTbls, Scheme^.num_messages);
  for i:= 0 to Scheme^.num_messages - 1 do
  begin
    if (i = 0) then MsgDesc:= Scheme^.messages^ else
    MsgDesc:= MsgDesc.next^;
    FStrms[str_idx].StrTbls[i].Name:= string(MsgDesc.name);
    FStrms[str_idx].StrTbls[i].idx:= i;
    SetLength(FStrms[str_idx].StrTbls[i].FDesc, MsgDesc.num_fields);
    for j:= 0 to MsgDesc.num_fields - 1 do
    begin
      if (j = 0) then FieldDesc:= MsgDesc.fields^ else
        FieldDesc:= FieldDesc.next^;
      FStrms[str_idx].StrTbls[i].FDesc[j].name:= string(FieldDesc.name);
      FStrms[str_idx].StrTbls[i].FDesc[j].field_type:= string(FieldDesc.field_type);
      FStrms[str_idx].StrTbls[i].FDesc[j].size:= FieldDesc.size;
      FStrms[str_idx].StrTbls[i].FDesc[j].offset:= FieldDesc.offset;
    end;
  end;
end;

procedure TCGCoreStorage.DecodeScheme;
var
  i, cnt: integer;
  is_found: boolean;
begin
  CgCore.Mutex.Lock;
  try
//--- Check scheme type
    if (Scheme^.scheme_type = CG_SCHEME_BINARY) then //Currently only binary supported
    begin
//--- Check for stream exists
      is_found:= false;
      for i:= 0 to length(FStrms) - 1 do
      begin
        if (FStrms[i].Name = CurLstnr.Name) then
        begin
          if (length(FStrms[i].StrTbls) = 0) then
            DecodeMessages(Scheme, i); //Decode messages
          is_found:= true;
          break;
        end;
      end;
//--- Build new streams tables
      if (not is_found) then
      begin
//--- Create new streams
        cnt:= Length(FStrms);
        SetLength(FStrms, Cnt + 1);
        FStrms[cnt].Name:= CurLstnr.Name;
//--- Decode messages
        DecodeMessages(Scheme, cnt);
      end;
    end;
  finally
    CgCore.Mutex.Unlock;
  end;
end;

//--- fill stream data
procedure TCGCoreStorage.FillStrData;
var
  StrData: tcg_msg_streamdata;
  buff: array of byte;
  i, j, k, w, s_idx, t_idx: integer;
  is_found: boolean;
  iBuff: array[0..7] of byte;
  replId: int64_t;
begin
  CgCore.Mutex.Lock;
  try
    if(msg^.msg_type = CG_MSG_STREAM_DATA) then
    begin
      StrData:= pcg_msg_streamdata(msg)^;
//--- Get data size
      if(StrData.data_size > 0) then
      begin
//--- Get data
        Setlength(buff, StrData.data_size);
        Move(StrData.data^, Pointer(buff)^, StrData.data_size);
//--- Check for new data (use var replId for relpAct)
        Move(buff[15], iBuff, Sizeof(iBuff));
        replId:= BuffToInt64(iBuff);
        if (replId = 0) then         //relpAct
        begin
//--- Get replID
          Move(buff[0], iBuff, Sizeof(iBuff));
          replId:= BuffToInt64(iBuff);
//--- Find exist table
          is_found:= false;
          s_idx:= -1;
          t_idx:= -1;
          for i:= 0 to Length(FStrms) - 1 do
          begin
            if (CurListn.Name = FStrms[i].Name) then  // stream found
            begin
              s_idx:= i;
              for j:= 0 to Length(FStrms[i].StrTbls) - 1 do
              begin
                if (StrData.msg_index = FStrms[i].StrTbls[j].idx) then // table found
                begin
                  t_idx:= j;
                 (* if (string(StrData.msg_name) = 'orders_aggr') then
                  begin
                    if(length(FStrms[i].StrTbls[j].Tbls) > 0) then
                    begin
                      for k := 0 to length(FStrms[i].StrTbls[j].Tbls) - 1 do
                      begin
                        if((FStrms[i].StrTbls[j].Tbls[k].tbl_rev <= StrData.rev) and
                           (FStrms[i].StrTbls[j].Tbls[k].repl_id = StrData.)) then
                        begin
                          for w:= 0 to length(FStrms[i].StrTbls[j].Tbls[k].Elements) - 1 do
                            Move(buff[FStrms[i].StrTbls[j].Tbls[0].Elements[w].offset],
                                 Pointer(FStrms[i].StrTbls[j].Tbls[0].Elements[w].Data.Buff)^,
                                 FStrms[i].StrTbls[j].Tbls[0].Elements[w].size);
                                 FStrms[i].StrTbls[j].Tbls[0].tbl_rev:= StrData.rev;
                          is_found:= true;
                          break;
                        end;
                      end;
                    end;
                  end else *)
//--- heartbeat move in to one table
                  if (string(StrData.msg_name) = 'heartbeat') then
                  begin
                    if (length(FStrms[i].StrTbls[j].Tbls) > 0) then
                    begin
                      for w:= 0 to length(FStrms[i].StrTbls[j].Tbls[0].Elements) - 1 do
                        Move(buff[FStrms[i].StrTbls[j].Tbls[0].Elements[w].offset],
                             Pointer(FStrms[i].StrTbls[j].Tbls[0].Elements[w].Data.Buff)^,
                             FStrms[i].StrTbls[j].Tbls[0].Elements[w].size);
                            FStrms[i].StrTbls[j].Tbls[0].tbl_rev:= StrData.rev;
                      is_found:= true;
                      break;
                    end;
                  end else //
                  begin
                    for k:= 0 to Length(FStrms[i].StrTbls[j].Tbls) - 1 do
                    begin
                      if (replId = FStrms[i].StrTbls[j].Tbls[k].repl_id) then
                      begin
                        for w:= 0 to length(FStrms[i].StrTbls[j].Tbls[k].Elements) - 1 do
                            Move(buff[FStrms[i].StrTbls[j].Tbls[k].Elements[w].offset],
                                 Pointer(FStrms[i].StrTbls[j].Tbls[k].Elements[w].Data.Buff)^,
                                 FStrms[i].StrTbls[j].Tbls[k].Elements[w].size);
                          FStrms[i].StrTbls[j].Tbls[k].tbl_rev:= StrData.rev;
                        is_found:= true;
                        break;
                      end;
                    end;
                  end;
                  if (is_found) then break;
//---- update orders_aggr
               (*   if (string(StrData.msg_name) = 'orders_aggr') then
                  begin
                    if (length(FStrms[i].StrTbls[j].Tbls) > 0) then
                    begin
                      for w:= 0 to length(FStrms[i].StrTbls[j].Tbls[0].Elements) - 1 do
                        Move(buff[FStrms[i].StrTbls[j].Tbls[0].Elements[w].offset],
                             Pointer(FStrms[i].StrTbls[j].Tbls[0].Elements[w].Data.Buff)^,
                             FStrms[i].StrTbls[j].Tbls[0].Elements[w].size);
                            FStrms[i].StrTbls[j].Tbls[0].tbl_rev:= StrData.rev;
                      is_found:= true;
                      break;
                    end;
                  end else
                  begin
                    for k:= 0 to Length(FStrms[i].StrTbls[j].Tbls) - 1 do
                    begin
                      if (replId = FStrms[i].StrTbls[j].Tbls[k].repl_id) then
                      begin
                        for w:= 0 to length(FStrms[i].StrTbls[j].Tbls[k].Elements) - 1 do
                            Move(buff[FStrms[i].StrTbls[j].Tbls[k].Elements[w].offset],
                                 Pointer(FStrms[i].StrTbls[j].Tbls[k].Elements[w].Data.Buff)^,
                                 FStrms[i].StrTbls[j].Tbls[k].Elements[w].size);
                          FStrms[i].StrTbls[j].Tbls[k].tbl_rev:= StrData.rev;
                        is_found:= true;
                        break;
                      end;
                    end;
                  end;
                  if (is_found) then break;*)
                end;
              end;
              if (is_found) then break;
            end;
          end;
//--- add new table
          if (not is_found) then
          begin
            if((s_idx > -1) and (t_idx > -1)) then
            begin
              i:= Length(FStrms[s_idx].StrTbls[t_idx].Tbls);
              SetLength(FStrms[s_idx].StrTbls[t_idx].Tbls, i + 1);
              FStrms[s_idx].StrTbls[t_idx].Tbls[i].repl_id:= replId;
              FStrms[s_idx].StrTbls[t_idx].Tbls[i].tbl_idx:= FStrms[s_idx].StrTbls[t_idx].idx;
              FStrms[s_idx].StrTbls[t_idx].Tbls[i].tbl_rev:= StrData.rev;
              FStrms[s_idx].StrTbls[t_idx].Tbls[i].have_Isin:= false;
              FStrms[s_idx].StrTbls[t_idx].Tbls[i].have_clcode:= false;
              FStrms[s_idx].StrTbls[t_idx].Tbls[i].Selected:= false;
              FStrms[s_idx].StrTbls[t_idx].Tbls[i].IsBook:= false;
              FStrms[s_idx].StrTbls[t_idx].Tbls[i].ses_id:= -1;
//--- Fill table
              SetLength(FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements,
                                  length(FStrms[s_idx].StrTbls[t_idx].FDesc));
              for j:= 0 to Length(FStrms[s_idx].StrTbls[t_idx].FDesc) - 1 do
              begin
                FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].Field:= string(FStrms[s_idx].StrTbls[t_idx].FDesc[j].name);
                FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].offset:= FStrms[s_idx].StrTbls[t_idx].FDesc[j].offset;
                FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].size:= FStrms[s_idx].StrTbls[t_idx].FDesc[j].size;
                FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].Data.FieldType:= string(FStrms[s_idx].StrTbls[t_idx].FDesc[j].field_type);
                SetLength(FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].Data.Buff, FStrms[s_idx].StrTbls[t_idx].FDesc[j].size);
                Move(buff[FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].offset],
                     Pointer(FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].Data.Buff)^,
                     FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].size);
                if (string(FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].Field) = 'sess_id') then
                  FStrms[s_idx].StrTbls[t_idx].Tbls[i].ses_id:=
                    BuffToInt64(FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].Data.Buff);
                if (string(FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].Field) = 'isin_id') then
                  FStrms[s_idx].StrTbls[t_idx].Tbls[i].have_Isin:= true;
                if (string(FStrms[s_idx].StrTbls[t_idx].Tbls[i].Elements[j].Field) = 'client_code') then
                  FStrms[s_idx].StrTbls[t_idx].Tbls[i].have_clcode:= true;
              end;
//--- if new session table, then build session
            if ((FStrms[s_idx].StrTbls[t_idx].Name = 'session') and
                (FStrms[s_idx].Name = 'FORTS_FUTINFO_REPL')) then
              BuildSession(s_idx, t_idx);
            end;
          end;
        end;
      end;
    end;
  finally
    CgCore.Mutex.Unlock;
  end;
end;

//--- delete tables
procedure TCGCoreStorage.DeleteTables;
var
  i: integer;
begin
  CgCore.Mutex.Lock;
  try
//--- Clear storage stream
    if (Length(FStrms) > 0) then
    begin
      for i:= 0 to Length(FStrms) - 1 do
      begin
        if (aListener.Name = FStrms[i].Name) then
        begin
          SetLength(FStrms[i].StrTbls, 0);
          break;
        end;
      end;
    end;
  finally
    CgCore.Mutex.Unlock;
  end;
end;

//--- Clear storage tables
procedure TCGCoreStorage.ClearTables;
var
  i, j, k, w, cnt: integer;
  is_found: boolean;
begin
  CgCore.Mutex.Lock;
  try
    if (Length(FStrms) > 0) then
    begin
      is_found:= false;
      for i:= 0 to Length(FStrms) - 1 do
      begin
        if (CListnr.Name = FStrms[i].Name) then
        begin
          for j:= 0 to Length(FStrms[i].StrTbls) - 1 do
          begin
            k:= 0;
            while k < Length(FStrms[i].StrTbls[j].Tbls) do
            begin
              if (FStrms[i].StrTbls[j].Tbls[k].tbl_idx = data^.table_idx) then
              begin
                if (FStrms[i].StrTbls[j].Tbls[k].tbl_rev < data^.table_rev) then
                 begin
                   for w:= k to Length(FStrms[i].StrTbls[j].Tbls) - 2 do
                      CopyTable(FStrms[i].StrTbls[j].Tbls[w + 1],
                                            FStrms[i].StrTbls[j].Tbls[w + 1]);
                   cnt:= Length(FStrms[i].StrTbls[j].Tbls);
                   SetLength(FStrms[i].StrTbls[j].Tbls, Cnt - 1);
                   dec(k);
                 end else
                  if (Data^.table_rev = CG_MAX_REVISON) then
                    FStrms[i].StrTbls[j].Tbls[k].tbl_rev:= 0;
              end;
              inc(k);
            end;
          end;
        end;
        break;
      end;
    end;
  finally
    CgCore.Mutex.Unlock;
  end;
end;

//--- buffer to int64
function TCGCoreStorage.BuffToInt64;
var
  i: integer;
begin
  if (sizeof(buff) > 8) then
  begin
    result:= -1;
    exit;
  end;
  result:= 0;
  for i := 0 to length(buff) - 1 do
    result:= result + (Buff[i] shl (i*8));
end;

//--- buffer to string
function TCGCoreStorage.BuffToString;
var
  i: integer;
begin
  result:= '';
  for i := 0 to length(buff) - 1 do
    result:= result + string(AnsiChar(buff[i]));
  result:= TrimRight(result);
end;

//--- copy table
procedure TCGCoreStorage.CopyTable;
var
  i, k: integer;
begin
  dest.repl_id:= source.repl_id;
  dest.tbl_idx:= source.tbl_idx;
  dest.tbl_rev:= source.tbl_rev;
  dest.ses_id:= source.ses_id;
  dest.have_Isin:= source.have_Isin;
  dest.have_clcode:= source.have_clcode;
  dest.Selected:= source.Selected;
  dest.IsBook:= source.IsBook;
  k:= Length(source.Elements);
  SetLength(dest.Elements, k);
  for i:= 0 to k - 1 do
  begin
    dest.Elements[i].Field:= source.Elements[i].Field;
    dest.Elements[i].offset:= source.Elements[i].offset;
    dest.Elements[i].size:= source.Elements[i].size;
    dest.Elements[i].Data.FieldType:= source.Elements[i].Data.FieldType;
    SetLength(dest.Elements[i].Data.Buff, dest.Elements[i].size);
    Move(source.Elements[i].Data.Buff[0], Pointer(dest.Elements[i].Data.Buff)^,
                                                         dest.Elements[i].size);
  end;
end;

procedure TCGCoreStorage.BuildSession;
var
  i: integer;
begin
  CgCore.Mutex.Lock;
  try
    i:= Length(FStrms[a_idx].StrTbls[b_idx].Tbls);
    if (i > 1) then
    begin
    //TODO: Check session ID ???????????????????????????????
      CopyTable(FStrms[a_idx].StrTbls[b_idx].Tbls[i-1],
                FStrms[a_idx].StrTbls[b_idx].Tbls[0]);
      SetLength(FStrms[a_idx].StrTbls[b_idx].Tbls, 1);
    end;
//--- Store session
    for i:= 0 to length(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements) - 1 do
    begin
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'sess_id') then
        FSymbInfo.CurSess.sess_id:= int32_t(BuffToInt64(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff)) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'begin') then
        FSymbInfo.CurSess.sess_begin:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'end') then
        FSymbInfo.CurSess.sess_end:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'state') then
        FSymbInfo.CurSess.state:= Integer(BuffToInt64(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff)) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'opt_sess_id') then
        FSymbInfo.CurSess.opt_sess_id:= int32_t(BuffToInt64(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff)) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'inter_cl_begin') then
        FSymbInfo.CurSess.inter_cl_begin:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'inter_cl_end') then
        FSymbInfo.CurSess.inter_cl_end:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'inter_cl_state') then
        FSymbInfo.CurSess.inter_cl_state:= Integer(BuffToInt64(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff)) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'eve_on') then
        FSymbInfo.CurSess.eve_on:= BuffToInt8(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'eve_begin') then
        FSymbInfo.CurSess.eve_begin:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'eve_end') then
        FSymbInfo.CurSess.eve_end:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'mon_on') then
        FSymbInfo.CurSess.mon_on:= BuffToInt8(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'mon_begin') then
        FSymbInfo.CurSess.mon_begin:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'mon_end') then
        FSymbInfo.CurSess.mon_end:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'pos_transfer_begin') then
        FSymbInfo.CurSess.pos_transfer_begin:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
      if (FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Field = 'pos_transfer_end') then
        FSymbInfo.CurSess.pos_transfer_end:= BuffToTime(FStrms[a_idx].StrTbls[b_idx].Tbls[0].Elements[i].Data.Buff) else
    end;
  finally
    CgCore.Mutex.Unlock;
  end;
//--- Update symbols
  UpdateSymbols();
end;

function TCGCoreStorage.BuffToTime;
begin
  if (Length(buff) <>  10) then exit;
  result.year:= buff[0] + (buff[1] shl 8);
  result.month:= buff[2];
  result.day:= buff[3];
  Result.hour:= buff[4];
  Result.minute:= buff[5];
  Result.second:= buff[6];
  result.msec:= buff[7] + (buff[8] shl 8);
end;

function TCGCoreStorage.BuffToInt8;
begin
  result:= 0;
  if (length(buff) <> 1) then exit;
  result:= buff[0];
end;

procedure TCGCoreStorage.UpdateSymbols;
var
  i, j, k, w, z: integer;
  no_opt_upd, no_fut_upd: boolean;
begin
  CgCore.Mutex.Lock;
  try
    no_opt_upd:= false;
    no_fut_upd:= false;
    if (FSymbInfo.CurSess.sess_id = -1) then exit;
    if (FSymbInfo.CurSess.opt_sess_id = -1) then exit;
    if (FSymbInfo.CurSess.sess_id = FSymbInfo.Fut.sess_id) then no_fut_upd:= true;
    if (FSymbInfo.CurSess.opt_sess_id = FSymbInfo.Opt.sess_id) then no_opt_upd:= true;
    for i:= 0 to length(FStrms) - 1 do
    begin
      if (FStrms[i].Name = 'FORTS_FUTINFO_REPL') then
      begin
        if (no_fut_upd = true) then Continue;
        SetLength(FSymbInfo.Fut.smbls, 0);
        for j:= 0 to length(FStrms[i].StrTbls) - 1 do
        begin
          if (FStrms[i].StrTbls[j].Name = 'fut_sess_contents') then
          begin
            for k:= 0 to length(FStrms[i].StrTbls[j].Tbls) - 1 do
            begin
              if (FStrms[i].StrTbls[j].Tbls[k].ses_id = FSymbInfo.CurSess.sess_id) then
              begin
                w:= Length(FSymbInfo.Fut.smbls);
                SetLength(FSymbInfo.Fut.smbls, w + 1);
                FSymbInfo.Fut.smbls[w].ResBook:= false;
                FSymbInfo.Fut.sess_id:= FSymbInfo.CurSess.sess_id;
                for z:= 0 to length(FStrms[i].StrTbls[j].Tbls[k].Elements) - 1 do
                begin
                  if (FStrms[i].StrTbls[j].Tbls[k].Elements[z].Field = 'isin_id') then
                  begin
                    FSymbInfo.Fut.smbls[w].isin_id:= uint32_t(BuffToInt64(FStrms[i].StrTbls[j].Tbls[k].Elements[z].Data.Buff));
                    FSymbInfo.Fut.smbls[w].StrNum:= i;
                    FSymbInfo.Fut.smbls[w].TblNum:= j;
                  end else
                  if (FStrms[i].StrTbls[j].Tbls[k].Elements[z].Field = 'isin') then
                    FSymbInfo.Fut.smbls[w].isin:= BuffToString(FStrms[i].StrTbls[j].Tbls[k].Elements[z].Data.Buff) else
                   if (FStrms[i].StrTbls[j].Tbls[k].Elements[z].Field = 'short_isin') then
                     FSymbInfo.Fut.smbls[w].sh_isin:= BuffToString(FStrms[i].StrTbls[j].Tbls[k].Elements[z].Data.Buff) else
                   if (FStrms[i].StrTbls[j].Tbls[k].Elements[z].Field = 'name') then
                     FSymbInfo.Fut.smbls[w].name:= BuffToString(FStrms[i].StrTbls[j].Tbls[k].Elements[z].Data.Buff);
                end;
              end;
            end;
            break;
          end;
        end;
      end else
      if (FStrms[i].Name = 'FORTS_OPTINFO_REPL') then
      begin
        if (no_opt_upd = true) then Continue;
        SetLength(FSymbInfo.Opt.smbls, 0);
        for j:= 0 to length(FStrms[i].StrTbls) - 1 do
        begin
          if (FStrms[i].StrTbls[j].Name = 'opt_sess_contents') then
          begin
            for k:= 0 to length(FStrms[i].StrTbls[j].Tbls) - 1 do
            begin
              if (FStrms[i].StrTbls[j].Tbls[k].ses_id = FSymbInfo.CurSess.opt_sess_id) then
              begin
                w:= Length(FSymbInfo.Opt.smbls);
                SetLength(FSymbInfo.Opt.smbls, w + 1);
                FSymbInfo.Opt.smbls[w].ResBook:= false;
                FSymbInfo.Opt.sess_id:= FSymbInfo.CurSess.opt_sess_id;
                for z:= 0 to length(FStrms[i].StrTbls[j].Tbls[k].Elements) - 1 do
                begin
                  if (FStrms[i].StrTbls[j].Tbls[k].Elements[z].Field = 'isin_id') then
                  begin
                    FSymbInfo.Opt.smbls[w].isin_id:= uint32_t(BuffToInt64(FStrms[i].StrTbls[j].Tbls[k].Elements[z].Data.Buff));
                    FSymbInfo.Opt.smbls[w].StrNum:= i;
                    FSymbInfo.Opt.smbls[w].TblNum:= j;
                  end else
                  if (FStrms[i].StrTbls[j].Tbls[k].Elements[z].Field = 'isin') then
                    FSymbInfo.Opt.smbls[w].isin:= BuffToString(FStrms[i].StrTbls[j].Tbls[k].Elements[z].Data.Buff) else
                   if (FStrms[i].StrTbls[j].Tbls[k].Elements[z].Field = 'short_isin') then
                     FSymbInfo.Opt.smbls[w].sh_isin:= BuffToString(FStrms[i].StrTbls[j].Tbls[k].Elements[z].Data.Buff) else
                   if (FStrms[i].StrTbls[j].Tbls[k].Elements[z].Field = 'name') then
                     FSymbInfo.Opt.smbls[w].name:= BuffToString(FStrms[i].StrTbls[j].Tbls[k].Elements[z].Data.Buff);
                end;
              end;
            end;
            break;
          end;
        end;
      end
    end;
  finally
    CgCore.Mutex.Unlock;
  end;
end;
//-- end of cgcore storage


end.
