unit Cgate_types;

interface

uses
  WinApi.Windows, System.SysUtils, cgate;

//--- Result codes
const
  MAX_INT = $7FFFFFFF;
  RES_ERROR = -1;
  RES_OK = 0;
  RES_STR_NOTFOUND = 1;
  RES_TABLE_NOTFOUND = 2;
  RES_ISIN_NOTFOUND = 3;
  RES_ISIN_ID_NOTFOUND = 4;

type
//--- incomming settings
  TOutSettings = packed record
    EnvStr: string;                      // Envitonment patameter
    ConnStr: string;                     // Connection patameter
    LstnrCnt: integer;                   // Repl Listeners count
    ListnrsNew: array of string;         // repl listeners Create parameters
    LOpenParam: array of string;         // repl listeners Open parameters
    Transactions: integer;               // Transactions count
  end;
//---
  TP2Set = packed record
    lpRegKey: PAnsiChar;                 // Cgate reg key
    lpRegValue: PAnsiChar;               // CGate reg value
    lpMachine: PWideChar;                // Machine local address
    lpServName: PWideChar;               // CGate service name
    lpSet: TOutSettings;                 // out Settings (in)
  end;

//--- Base storage element
  TBaseElement = packed record
    FieldType: string;                   // Data type
    Buff: array of byte;                 // Data Buffer
  end;

//--- Element
  TElement = packed record
    Field: string;                       // Table element name
    offset: size_t;                      // offset data
    size: size_t;                        // data size
    Data: TBaseElement;                  // Data
  end;

//--- Isin book
  PBook = ^TBook;
  TBook = packed record
    isin_id: size_t;                     // inin id
    sell_cnt: integer;                   //
    sell_price: array [0..49] of double; //
    sell_vol: array [0..49] of integer;  //
    buy_cnt: integer;                    //
    buy_price: array [0..49] of double;  //
    buy_vol: array [0..49] of integer;   //
  end;
// --- isin Option desk
  TOptionDesk = packed record            //TODO: ??????
    isin_id: size_t;                     //
  end;

//--- Table
  PTable = ^TTable;
  TTable = packed record
    repl_id: int64_t;                    // Repl ID
    tbl_idx: uint32_t;                   // Table ID
    tbl_rev: int64_t;                    // Table revision
    ses_id: int32_t;                     // Futures session ID
//    opt_ses_id: int32_t;                 // Options session ID
    have_Isin: boolean;                  // Have isisn ID
    have_clcode: boolean;                // Have client code
    Selected: boolean;                   // Symbol selected
    IsBook: boolean;                     // Get book select
//    ElCnt: size_t;                       // Table elements count
    Elements: array of TElement;         // Table data records
  end;

//--- Listener
  TListener = packed record
    Name: string;                        // Listener name
    StrDone: boolean;                    // Listener done CG_MSG_CLOSE resived
    OnLine: boolean;                     // Listener online
    CanClose: boolean;                   // Msg CG_MSG_DATA resived
    hLstnr: THandle;                     // Listener handle
    State: uint32_t;                     // Listener state
    OpenParam: string;                   // Open listener param
    LifeNum: uint32_t;                   // Listener LifeNum scheme
  end;

//--- Publisher
  TPublisher = packed record
    Name: string;                        // Publisher name
    hPub: THandle;
    State: uint32_t;
  end;

//--- Replication listeners
  TReplLisrteners = packed record
    Cnt: integer;                        // Repl listeners count
    Lstnrs: array of TListener;          // Repl listeners
  end;

//-- field description
  TDescription = packed record
    name: string;                        // Field name  May be NULL if not defined
    field_type: string;                  // Field type
    size: size_t;                        // Size in bytes
    offset: size_t;                      // Offset from beginning of data in bytes
   end;
//--- Main table
  TMainTable = packed record
    Name: string;                        // Table name
    idx: size_t;                         // Table index
    FDesc: array of TDescription;        // Table field desc
    Tbls: array of TTable;               // Child tables
  end;
//--- Cgate stogage streams
  TStorageStream = packed record
    Name: string;                        // Stream name
    StrTbls: array of TMainTable;        // Stream tables
  end;

//--- Symbol
  TSymbol = packed record
    ResBook: boolean;                    // Resived book
    isin_id: int32_t;                    // Isin ID
    isin: string;                        // Instrument isin
    sh_isin: string;                     // Short isin
    name: string;                        // Instrument full name
    StrNum: integer;                     // Stream number
    TblNum: integer;                     // Table nunber
   // ChNum: integer;                      // Child number
  end;

//--- Symbols
  TSymbols = packed record
    sess_id: int32_t;                    // Session id
    smbls: array of TSymbol;             // Symbols with same session id
  end;

// Session information
  TSession = packed record
    sess_id: int32_t;                    // Futures session ID
    sess_begin: tcg_time;                // Start session
    sess_end: tcg_time;                  // End Session
    state: int32_t;                      // Session state
    opt_sess_id: int32_t;                // Options session ID
    inter_cl_begin: tcg_time;            // Start cliring time
    inter_cl_end: tcg_time;              // End cliring time
    inter_cl_state: int32_t;             // State cliring
    eve_on: byte;                        // Evening session will be held
    eve_begin: tcg_time;                 // Evening session start time
    eve_end: tcg_time;                   // Evening session end time
    mon_on: byte;                        // Morning session will be held
    mon_begin: tcg_time;                 // Morning session start time
    mon_end: tcg_time;                   // Morning session end time
    pos_transfer_begin: tcg_time;        // Start of transfer interval
    pos_transfer_end: tcg_time;          // End of transfer interval
  end;

//--- Symbols information
  PSymbolsTradeInfo = ^TSymbolsTradeInfo;
  TSymbolsTradeInfo = packed record
    Fut: TSymbols;                       // Futures
    Opt: TSymbols;                       // Options
    CurSess: TSession;                   // Session information
  end;


implementation


end.
