unit Old_Cgate_classes;

interface

uses
  WinApi.Windows, System.Classes, System.SysUtils, cgate_types, cgate_functions,
  cgate;

type
//--- Mutex
  TMutex = class
    private
      FSection: TRTLCriticalSection;
    public
      constructor Create();
      destructor Destroy(); override;
      procedure Lock();
      procedure Unlock();
  end;

//---
  TTables = array of TTable;

//--- Cgate core Storage
  TCGCoreStorage = class
    private
      FMutex: TMutex;
      FProtTblCnt: integer;            // Prototype table count
      FProtTables: TTables;            // Prototype tables
      FTblCnt: integer;                // Tables count
      FTables: array of TTable;        // Tables
      FBooks: array of TBook;          // Books
      FOptDesk: array of TOptionDesk;  // Potion desks
    protected
      procedure DecodeScheme(Scheme: pcg_scheme_desc; CurThread: TThread);
      procedure FillSnapShotTables(Msg: pcg_msg; CurThread: TThread);
      procedure FillOnlLineTables(Msg: pcg_msg; CurThread: TThread);
      procedure ClearStorage();
      procedure ClearTable(idx: uint32_t);
      procedure CopyTable(const source: TTable; var target: TTable);
      function BuffToInt(const buff: array of byte): int64_t;
    public
      constructor Create();
      destructor Destroy(); override;
      property Mutex: TMutex read FMutex write FMutex;
    property TablesCnt: integer read FTblCnt;
    property ProtTables: TTables read FProtTables;
  end;

//--- P2Lsn thread
  TConnThread = class(TThread)
    private
      FName: AnsiString;           // Thraed name
      FConSet: AnsiString;         // Connection settings
      FLsnrSet: AnsiString;        // Listener settings
      FDoExit: boolean;            // Do Exit thread
      FDone: boolean;            // Thread done
      FConn: THandle;              // Connection handle
      FListener: THandle;          // Listener
      FPublisher: THandle;         // Publisher
      FConState: uint32_t;         // Connection state
      FLsnrState: uint32_t;         // Listener state
      FDelay: uint32_t;            // Delay of ...
      FOpenParam: PAnsiChar;       // Open Connection param
      FLifeNum: uint64_t;          // Lifenum scheme
    //  FPubState: uint32_t;        // Publisher state
      FIsOnline: boolean;          // Straem online mode
      FMsgType: uint32_t;          // Message tyte
    protected
      procedure Execute(); override;
      function MakeCallBack(const Param: ansistring): CG_RESULT;
      function MsgCallBack(hConn: THandle; hListener: THandle; msg: pcg_msg; Data: Pointer): CG_RESULT; stdcall;
    public
      constructor Create(cSet, lSet: AnsiString; TDelay: uint64_t; name: ansistring);
      destructor Destroy(); override;
      property LifeNum: uint64_t read FLifeNum;
      property Name: AnsiString read FName;
  end;
//--- connection manager
  TConnectionManager = class
    private
      FGlobSet: TP2Set;                // cgate common settings
      FConnsCnt: integer;              // connections count
      FConns: array of TConnThread;    // connections
    protected
      function GetStrName(const instr: string): string;
    public
      constructor Create(Sett: TP2Set);
      destructor Destroy(); override;
  end;
//--- Main class CGate core
  TCgate_core = class
    private
      FCgDllFPath: string;             // cgate dll full path
      FCgateHandle: THandle;           // cgate dll handle
      FConMgr: TConnectionManager;     // connection manager
      FCgEnvir: CG_RESULT;             // Enviroiment status
      FStorage: TCGCoreStorage;        // cgate core storage
    protected

    public
      constructor Create();
      destructor Destroy(); override;
      function Connect(CGSet: TP2Set): PAnsichar;
      function Disconnect(): boolean;
    property CgEnvir: CG_RESULT read FCgEnvir;
    property CgateHandle: THandle read FCgateHandle;
    property Storage: TCGCoreStorage read FStorage;
    property CgDllFPath: string read FCgDllFPath;
  end;

var
  CgCore: TCgate_core;

implementation

//--- Cgate core storage
constructor TCGCoreStorage.Create;
begin
  inherited Create;
  if(not Assigned(Mutex)) then FMutex:= TMutex.Create;
  FProtTblCnt:= 0;
  SetLength(FProtTables, 0);
  FTblCnt:= 0;
  SetLength(FTables, 0);
  SetLength(FBooks, 0);
  SetLength(FOptDesk, 0);
end;

destructor TCGCoreStorage.Destroy;
begin
  FProtTblCnt:= 0;
  SetLength(FProtTables, 0);
  FTblCnt:= 0;
  SetLength(FTables, 0);
  SetLength(FBooks, 0);
  SetLength(FOptDesk, 0);
//--- Destroy mutex
  if Assigned(FMutex) then FreeAndNil(FMutex);
  inherited Destroy;
end;

procedure TCGCoreStorage.DecodeScheme;
var
  i, j: integer;
  MsgDesc: tcg_message_desc;
  FieldDesc: tcg_field_desc;
  temp_tbl: TTable;
begin
  Mutex.Lock;
  try
//--- Check scheme type
  if (Scheme^.scheme_type = CG_SCHEME_BINARY) then //Currently only binary supported
  begin
//--- Decode messages
    for i:= 0 to Scheme^.num_messages - 1 do
    begin
      if (i = 0) then MsgDesc:= Scheme^.messages^ else
        MsgDesc:= MsgDesc.next^;
//--- Build temp table
      temp_tbl.tbl_idx:= i;
      temp_tbl.Name:= AnsiString(MsgDesc.name);
      temp_tbl.ElCnt:= MsgDesc.num_fields;
      SetLength(temp_tbl.Elements, temp_tbl.ElCnt);
      temp_tbl.have_isin:= false;
//--- Get fields
      for j:= 0 to temp_tbl.ElCnt - 1 do
      begin
        if (j = 0) then FieldDesc:= MsgDesc.fields^ else
          FieldDesc:= FieldDesc.next^;
        temp_tbl.Elements[j].Field:= AnsiString(FieldDesc.name);
        if(AnsiString(FieldDesc.name) = AnsiString('isin_id')) then
          temp_tbl.have_isin:= true;
        temp_tbl.Elements[j].offset:= FieldDesc.offset;
        temp_tbl.Elements[j].size:= FieldDesc.size;
        setlength(temp_tbl.Elements[j].Dat, FieldDesc.size);
        temp_tbl.Elements[j].FieldType:= AnsiString(FieldDesc.field_type);
      end;
//--- Add prototype table
      temp_tbl.PName:= TConnThread(CurThread).FName;
      temp_tbl.tbl_idx:= i; //table index
      FProtTblCnt:= FProtTblCnt + 1;
      SetLength(FProtTables, FProtTblCnt);
      FProtTables[FProtTblCnt - 1]:= temp_tbl;
      temp_tbl.ElCnt:= 0;
      SetLength(temp_tbl.Elements, 0);
    end;
  end;
  finally
    Mutex.Unlock;
  end;
end;

procedure TCGCoreStorage.FillSnapShotTables;
var
  StrData: pcg_msg_streamdata;
  buff: array of byte;
  buff_size: Int64;
  i: integer;
  CurTable: TTable;
  is_found: boolean;
begin
  Mutex.Lock;
  try
    if(msg^.msg_type = CG_MSG_STREAM_DATA) then
    begin
      StrData:= pcg_msg_streamdata(msg);
      buff_size:= StrData^.data_size;
      Setlength(buff, buff_size);
      Pointer(buff):= StrData^.data;
//--- Find prototype table
      is_found:= false;
      for i := 0 to FProtTblCnt - 1 do
      begin
//--- Copy prototype table
        CopyTable(ProtTables[i], CurTable);
        if((CurTable.PName = TConnThread(CurThread).FName) and
          (CurTable.name = StrData^.msg_name))  then
        begin
          is_found:= true;
          break;
        end;
      end;
      if (is_found) then
      begin
//--- Fill data
        for i := 0 to CurTable.ElCnt - 1 do
        begin
          Move(buff[CurTable.Elements[i].offset],
               Pointer(CurTable.Elements[i].Dat)^,
                       CurTable.Elements[i].size);
        end;
//--- Not old data
        if( BuffToInt(CurTable.Elements[2].Dat) = 0) then
         begin
//--- Find same table in woking tables
           is_found:= false;
           for i := 0 to TablesCnt - 1 do
           begin
             if(BuffToint(FTables[i].Elements[0].Dat) = BuffToint(CurTable.Elements[0].Dat)) then     //replD
             begin
               if(BuffToint(FTables[i].Elements[1].Dat) <> BuffToint(CurTable.Elements[1].Dat)) then  //RevID
                  FTables[i]:= CurTable;
               is_found:= true;
               break;
             end;
           end;
           if(not is_found) then
           begin
//--- Create new work table
             inc(FTblCnt);
             SetLength(FTables, FTblCnt);
             CopyTable(CurTable, FTables[FTblCnt - 1]);
             FTables[FTblCnt - 1].tbl_rev:= StrData^.rev;
           end;
         end;
      end;
    end;
  finally
    Mutex.Unlock;
  end;
end;

procedure TCGCoreStorage.FillOnlLineTables;
begin
  //TODO:
{  Mutex.Lock;
  try

  finally
    Mutex.Unlock;
  end;          }
end;

procedure TCGCoreStorage.ClearStorage;
begin
  Mutex.Lock;
  try
    FProtTblCnt:= 0;
    SetLength(FProtTables, FProtTblCnt);
    SetLength(FBooks, FProtTblCnt);
    SetLength(FOptDesk, FProtTblCnt);
    FTblCnt:= 0;
    SetLength(FTables, FTblCnt);
    SetLength(FBooks, FTblCnt);
    SetLength(FOptDesk, FTblCnt);
  finally
    Mutex.Unlock;
  end;
end;

procedure TCGCoreStorage.ClearTable;
begin
  //TODO:
{  Mutex.Lock;
  try

  finally
    Mutex.Unlock;
  end; }
end;

procedure TCGCoreStorage.CopyTable;
var
  i: integer;
begin
  target.PName:= source.PName;
  target.tbl_idx:= source.tbl_idx;
  target.tbl_rev:= source.tbl_rev;
  target.have_isin:= source.have_isin;
  target.Name:= source.Name;
  target.ElCnt:= source.ElCnt;
  SetLength(target.Elements, target.ElCnt);
  for i := 0 to target.ElCnt - 1 do
  begin
    target.Elements[i].Field:= source.Elements[i].Field;
    target.Elements[i].offset:= source.Elements[i].offset;
    target.Elements[i].size:= source.Elements[i].size;
    target.Elements[i].FieldType:= source.Elements[i].FieldType;
    SetLength(target.Elements[i].Dat, target.Elements[i].size);
    Move(source.Elements[i].Dat, Pointer(target.Elements[i].Dat)^, target.Elements[i].size)
  end;
end;

function TCGCoreStorage.BuffToInt;
var
  i, size: integer;
begin
  result:= 0;
  size:= Length(buff);
  for i := 0 to size - 1 do
    result:= result + (Buff[i] shl (i*8));
end;
//--- End Cgate core storage



//--- connection Manager ---

function TConnectionManager.GetStrName;
var
  k: integer;
begin
  result:= instr;
  k:= pos('lsnr_', result);
  if (k > 0) then Delete(result, 1, k + 4);
end;


constructor TConnectionManager.Create;
var
  i: integer;
  s: string;
begin
  inherited Create();
  FGlobSet:= Sett;
  FConnsCnt:= FGlobSet.lpConns.StrCnt;
  SetLength(FConns, FConnsCnt);
//--- create connections
  if(CGCore.CgEnvir = CG_ERR_OK) then
  begin
  for i := 0 to FConnsCnt - 1 do
  begin
    s:= GetStrName(FGlobSet.lpConns.LsnrSet[i]);
    FConns[i]:= TConnThread.Create(ansistring(FGlobSet.lpConns.StrSet[i]),
                                   ansistring(FGlobSet.lpConns.LsnrSet[i]),
                                   FGlobSet.lpConns.DelaySet[i],
                                   AnsiString(s));
     // FConns[i].Start;
    end;
  end;
end;

destructor TConnectionManager.Destroy;
var
  i: Integer;
begin
  for i:= 0 to FConnsCnt - 1 do
  begin
    if (not FConns[i].Terminated) then
    begin
      FConns[i].FDoExit:= true;
      FConns[i].Terminate;
      FConns[i].WaitFor;
      if (FConns[i].Terminated) then
        FreeAndNil(FConns[i]);
    end;
  end;
  FConnsCnt:= 0;
  inherited Destroy;
end;
//--- End Manager ---

//--- Mutex ---
constructor TMutex.Create;
begin
  InitializeCriticalSection(FSection);
  inherited Create();
end;

destructor TMutex.Destroy;
begin
  DeleteCriticalSection(FSection);
  inherited Destroy;
end;

procedure TMutex.Lock();
begin
  EnterCriticalSection(FSection);
end;

procedure TMutex.Unlock();
begin
  LeaveCriticalSection(FSection);
end;
//--- End Mutex ---

//--- Cgate_core -------
constructor TCgate_core.Create;
begin
  inherited Create;
  FCgEnvir:= CG_ERR_NULL;
  FStorage:= TCGCoreStorage.Create;
end;

destructor TCgate_core.Destroy;
begin
//--- Destroy connection manager, if exist and close enviroiment, if open
  Disconnect();
//--- Unload cgate library
  if(FCgateHandle <> 0) then
    if(FreeLibrary(FCgateHandle)) then FCgateHandle := 0;
//--- Destroy storage
  if(Assigned(FStorage)) then FreeAndNil(FStorage);
  inherited Destroy;
end;

function TCgate_core.Connect;
var
  env_string: PAnsiChar;
begin
  result:= nil;
//--- Check connection manager
  if(not Assigned(FConMgr)) then
  begin
//--- Check CGATE servide running
    if (GetCgService(CGSet.lpMachine, CGSet.lpServName)) then
    begin
//--- Get cgate dll full path
     FCgDllFPath:= GetCgDllFullPath(string(CGSet.lpRegKey),
                                    string(CGSet.lpRegValue),
                                    string(CGSet.lpCgDllName));
      if(CgDllFPath <> '') then
      begin
//--- Load cgate library
        FCgateHandle:= LoadCgLib(CgDllFPath);
        if(FCgateHandle <> 0) then
        begin
          env_string:= PansiChar(RawByteString(CGSet.lpConns.envStr));
          FCgEnvir:= cg_env_open(env_string);
          if(FCgEnvir = CG_ERR_OK) then
          begin
            if(not Assigned(FConMgr)) then FConMgr:= TConnectionManager.Create(CGSet);
            if(Assigned(FConMgr)) then
              Result:= PAnsiChar(GetCgVersion(PWideChar(CgDllFPath))) else
                Result:= PAnsiChar('Error: Connection manager not created!');
          end else Result:= PAnsiChar('Error: Cgate enviroiment not open!');
        end else Result:= PAnsiChar('Error: Cgate library not loaded!');
      end else Result:= PAnsiChar('Error: Get cgate library path failed!');
    end else Result:= PAnsiChar('Error: Spectra Gate service not running!');
  end else
    if(CgDllFPath <> '') then Result:= PAnsiChar(GetCgVersion(PWideChar(CgDllFPath))) else
      Result:= PAnsiChar('Error: Unable to find dll path!');
end;

function TCgate_core.Disconnect;
begin
  result:= false;
//--- Destroy connection manager
  if(Assigned(FConMgr)) then FreeAndNil(FConMgr);
  if(not Assigned(FConMgr)) then
  begin
//--- Close enviroiment
    if(FCgEnvir = CG_ERR_OK) then
    begin
      FCgEnvir:= cg_env_close();
      if(FCgEnvir = CG_ERR_OK) then
      begin
        FCgEnvir:= CG_ERR_NULL;
        result:= true;
      end;
    end else result:= true;
  end;
end;
//--- End Cgate core ---

//--- P2Lsn thread
constructor TConnThread.Create;
begin
  inherited Create(false);
  FName:= name;
  FConn:= 0;
  FListener:= 0;
  FPublisher:= 0;
  FOpenParam:= nil;
  FConState:= CG_STATE_CLOSED;
  FLsnrState:= CG_STATE_CLOSED;
  //FPubState:= CG_STATE_CLOSED;
//  FCanClose:= false;
  FIsOnline:= false;
  FLifeNum:= 0;
  FDoExit:= false;
  FDone:= false;
  FConSet:= cSet;
  FLsnrSet:= lSet;
  FDelay:= TDelay;
  FreeOnTerminate:= false;
end;

destructor TConnThread.Destroy;
begin
  inherited Destroy;
end;

//--- P2 msg CallBack
function TConnThread.MsgCallBack;
var
  h_res: CG_RESULT;
  Scheme: pcg_scheme_desc;
  LifeNum: pcg_data_lifenum;
  ClrDel: pcg_data_cleardeleted;
begin
//--- Recive messages
  case Msg^.msg_type of
    CG_MSG_OPEN:
      begin
        TConnThread(CurrentThread).FMsgType:= Msg^.msg_type;
        Scheme:= nil;
        h_res:= cg_lsn_getscheme(TConnThread(CurrentThread).FListener, Scheme);
        if(h_res = CG_ERR_OK) then CGCore.Storage.DecodeScheme(Scheme, CurrentThread);
      end;
    CG_MSG_CLOSE: TConnThread(CurrentThread).FIsOnline:= false;
    CG_MSG_TIME:;
    CG_MSG_DATA:; //    ()
    CG_MSG_STREAM_DATA: if(Msg^.data_size > 0) then
                          if(TConnThread(CurrentThread).FIsOnline) then
                            CGCore.Storage.FillOnlLineTables(msg, CurrentThread) else
                           CGCore.Storage.FillSnapShotTables(msg, CurrentThread);
    CG_MSG_DATAARRAY:;
    CG_MSG_OBJFAILED:;
    CG_MSG_PARSEERR:;
    CG_MSG_TN_BEGIN:;
    CG_MSG_TN_COMMIT:;
    CG_MSG_TN_ROLLBACK:;
    CG_MSG_P2MQ_RANGE_START:;
    CG_MSG_P2MQ_TIMEOUT:;
    CG_MSG_P2REPL_RANGE_START:;
    CG_MSG_P2REPL_LIFENUM:
      begin
        LifeNum:= pcg_data_lifenum(Msg^.data);
        if(TConnThread(CurrentThread).FLifeNum <> LifeNum^.life_number) then
        begin
          TConnThread(CurrentThread).FLifeNum:= LifeNum^.life_number;
          if(LifeNum^.flags <> CG_LIFENUM_CHANGE__DONT_CLEAR_STORAGE) then
            CgCore.Storage.ClearStorage();
        end;
      end;
    CG_MSG_P2REPL_CLEARDELETED:
      begin
        ClrDel:= pcg_data_cleardeleted(Msg^.data);
        if(ClrDel^.table_rev = CG_MAX_REVISON) then
          CgCore.Storage.ClearTable(ClrDel^.table_idx);
      end;
    CG_MSG_P2REPL_ONLINE: TConnThread(CurrentThread).FIsOnline:= true;
    CG_MSG_P2REPL_REPLSTATE:
      {if (Msg^.data_size > 0) then FOpenParam:= Msg^.data};
  end;
  result:= CG_ERR_OK;
end;

//---Call back
function P2Lsn_CallBack(hConn: THandle; hListener: THandle; msg: pcg_msg; Data: Pointer): CG_RESULT; stdcall;
begin
  Result:= TConnThread(Data).MsgCallBack(hConn, hListener, msg, Data);
end;

function TConnThread.MakeCallBack;
begin
  Result:= cg_lsn_new(FConn, PAnsiChar(Param), @P2Lsn_CallBack, nil, FListener);
end;

//--- P2Lsn thread
procedure TConnThread.Execute;
var
  h_res: CG_RESULT;
begin
  while (not FDone) do
  begin
    if (not FDoExit) then
    begin
     if (FConn = 0) then cg_conn_new(PAnsiChar(FConSet), FConn) else
      begin
        if((cg_conn_getstate(FConn, FConState)) = CG_ERR_OK) then
        begin
          case FConState of
            CG_STATE_ERROR:
              begin
                h_res:= cg_conn_close(FConn);
                if(h_res <> CG_ERR_OK) then
                  cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Error: connection not closed: %d',
                                                                      [h_res]))));
              end;
            CG_STATE_OPENING: ; //TODO:
            CG_STATE_CLOSED:
              begin
                h_res:= cg_conn_open(FConn, nil);
                if(h_res <> CG_ERR_OK) then
                  cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Error: connection not open: %d',
                                                                      [h_res]))));
              end;
            CG_STATE_ACTIVE:
              begin
                h_res:= cg_conn_process(FConn, FDelay, nil);
                if((h_res <> CG_ERR_OK) and (h_res <> CG_ERR_TIMEOUT)) then
                 cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Warning: connection state request: %d',
                                                               [h_res]))));
                if (FListener = 0) then
                begin
                  if (MakeCallBack(FLsnrSet) <> CG_ERR_OK) then
                    cg_log_infostr(PAnsiChar(AnsiString('--->>> Error: Listener not set callback! ')));
                end else
                begin
                  if(cg_lsn_getstate(FListener, FLsnrState) = CG_ERR_OK) then
                  begin
                    case FLsnrState of
                      CG_STATE_ERROR:
                        begin
                          cg_log_infostr(PAnsiChar(AnsiString('--->>> Error: Listener state error! ')));
                          h_res:= cg_lsn_close(FListener);
                          if(h_res <> CG_ERR_OK) then
                              cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Error: listener not closed: %d',
                                                             [h_res]))));
                        end;
                      CG_STATE_CLOSED:
                        begin
                          h_res:= cg_lsn_open(FListener, FOpenParam);
                          if(h_res <> CG_ERR_OK) then
                            cg_log_infostr(PAnsiChar(AnsiString(format('--->>> Error: listener not open: %d',
                                                                  [h_res]))));
                        end;
                      CG_STATE_ACTIVE:
                        begin
                          //---
                        end;
                      CG_STATE_OPENING:;
                    end;
                  end;
                end;
              end;
          end;
        end;
      end;
    end else
    begin
//--- close listener connection
      if (Fconn <> 0) then
      begin
        cg_log_infostr(PAnsiChar(AnsiString('--->>> Msg type = ' +
                                              AnsiString(inttoStr(FMsgType)))));
        if (FMsgType = CG_MSG_CLOSE) then;
        if (FListener <> 0) then
        begin
          if(cg_lsn_getstate(FListener, FLsnrState) = CG_ERR_OK) then
          begin
            cg_log_infostr(PAnsiChar(AnsiString('--->>> Listener get state. Name = ' + FName + ' state = ') +
                                             AnsiString(IntToStr(FLsnrState))));
            case FLsnrState of
             // CG_STATE_OPENING:;
              CG_STATE_ERROR,
              CG_STATE_ACTIVE:
                  if(cg_lsn_close(FListener) = CG_ERR_OK) then
                    if (cg_lsn_destroy(FListener) = CG_ERR_OK) then FListener:= 0;
              CG_STATE_CLOSED:
                  if (cg_lsn_destroy(FListener) = CG_ERR_OK) then FListener:= 0;
            end;
          end;
        end else
        begin
          if((cg_conn_getstate(FConn, FConState)) = CG_ERR_OK) then
          begin
            cg_log_infostr(PAnsiChar(AnsiString('--->>> Connection get state. Name = ' + FName + ' state = ') +
                                             AnsiString(IntToStr(FConState))));
            case FConState of
              //CG_STATE_OPENING,
              CG_STATE_ERROR,
              CG_STATE_ACTIVE:
//---  lsn_close  conn_close      CG_MSG_OPEN. (SDK) ???
                if (FMsgType = CG_MSG_OPEN) then
                  if (cg_conn_close(FConn)= CG_ERR_OK) then
                    if (cg_conn_destroy(FConn) = CG_ERR_OK) then FConn:= 0;
              CG_STATE_CLOSED:
                if (cg_conn_destroy(FConn) = CG_ERR_OK) then FConn:= 0;
             // CG_STATE_OPENING: ;
            end;
          end;
        end;
      end;
//--- Check
      if ((FConn = 0) and (FListener = 0) and (FPublisher = 0)) then FDone:= true;
    end;
  end;
//--- Thread done
end;
//--- End of connection thread


end.
