{
/*
        Copyright (c) 2012, MICEX-RTS. All rights reserved.

        Plaza-2 Client Gate API Description. Included are types,
        constants and functions related to the API.
        Contents are subject to change.

        All the software and documentation included in this and any
        other MICEX-RTS CGate Releasese is copyrighted by MICEX-RTS.

        Redistribution and use in source and binary forms, with or without
        modification, are permitted only by the terms of a valid
        software license agreement with MICEX-RTS.

        THIS SOFTWARE IS PROVIDED "AS IS" AND MICEX-RTS DISCLAIMS ALL WARRANTIES
        EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
        NON-INFRINGEMENT, MERCHANTABILITY OR FITNESS FOR A PARTICULAR
        PURPOSE.  MICEX-RTS DOES NOT WARRANT THAT USE OF THE SOFTWARE WILL BE
        UNINTERRUPTED OR ERROR-FREE.  MICEX-RTS SHALL NOT, UNDER ANY CIRCUMSTANCES, BE
        LIABLE TO LICENSEE FOR LOST PROFITS, CONSEQUENTIAL, INCIDENTAL, SPECIAL OR
        INDIRECT DAMAGES ARISING OUT OF OR RELATED TO THIS AGREEMENT OR THE
        TRANSACTIONS CONTEMPLATED HEREUNDER, EVEN IF MICEX-RTS HAS BEEN APPRISED OF
        THE LIKELIHOOD OF SUCH DAMAGES.
*/
}

unit cgate;

interface

uses
  WinApi.Windows;

//--- Variables ---
type
  int8_t = ShortInt;
  int16_t = SmallInt;
  int32_t = Integer;
  int64_t = Int64;
  uint8_t = Byte;
  uint8_tp = ^uint8_t;
  uint16_t = Word;
  uint32_t = LongWord;
  uint64_t = uInt64;

// ------------------------------------------------------------------
//--- API setup
const
  CG_VERSION_MAJOR = 5;
  CG_VERSION_MINOR = 3;
  CG_VERSION_PATCH = 5;
  CG_VERSION = (CG_VERSION_MAJOR shl 24) + (CG_VERSION_MINOR shl 16) + CG_VERSION_PATCH;
// ------------------------------------------------------------------
//-- enumerations

//--- error codes
  CG_RANGE_BEGIN         = $20000;
  CG_ERR_OK              = 0;
	CG_ERR_INTERNAL        = CG_RANGE_BEGIN;
	CG_ERR_INVALIDARGUMENT = 1 + CG_RANGE_BEGIN;
	CG_ERR_UNSUPPORTED     = 2 + CG_RANGE_BEGIN;
	CG_ERR_TIMEOUT         = 3 + CG_RANGE_BEGIN;
	CG_ERR_MORE            = 4 + CG_RANGE_BEGIN;
	CG_ERR_INCORRECTSTATE  = 5 + CG_RANGE_BEGIN;
	CG_ERR_DUPLICATEID     = 6 + CG_RANGE_BEGIN;
	CG_ERR_BUFFERTOOSMALL  = 7 + CG_RANGE_BEGIN;
	CG_ERR_OVERFLOW        = 8 + CG_RANGE_BEGIN;
	CG_ERR_UNDERFLOW       = 9 + CG_RANGE_BEGIN;
	CG_ERR_EOF             = 10 + CG_RANGE_BEGIN;
	CG_RANGE_END           = 11 + CG_RANGE_BEGIN;
//--- MY ERROR CODE
  CG_ERR_NULL            = 111;

//--- scheme types
  CG_SCHEME_BINARY                     = 1;

//---
  CG_SCHEME_BIN_MSG_HAS_ID					   = $00000001;
  CG_SCHEME_BIN_MSG_HAS_NAME					 = $00000002;
  CG_SCHEME_BIN_MSG_HAS_DESC					 = $00000004;
  CG_SCHEME_BIN_MSG_HAS_INDICES				 = $00000008;
  CG_SCHEME_BIN_MSG_HAS_TIMESTAMP_HINT = $00000010;
  CG_SCHEME_BIN_MSG_HAS_USER_HINT			 = $00000020;
  CG_SCHEME_BIN_FIELD_HAS_ID					 = $00000100;
  CG_SCHEME_BIN_FIELD_HAS_NAME				 = $00000200;
  CG_SCHEME_BIN_FIELD_HAS_DESC				 = $00000400;
  CG_SCHEME_BIN_FIELD_HAS_TYPE				 = $00000800;
  CG_SCHEME_BIN_FIELD_HAS_DEFVAL			 = $00001000;

//--- states
  CG_STATE_CLOSED	 = 0;
  CG_STATE_ERROR	 = 1;
  CG_STATE_OPENING = 2;
  CG_STATE_ACTIVE	 = 3;

//--- message creation flags
  CG_KEY_INDEX = 0;
  CG_KEY_ID		 = 1;
  CG_KEY_NAME	 = 2;

// listener flags

//--- publisher flags
  CG_PUB_NEEDREPLY = $1;

//--- message types
  CG_MSG_OPEN			           = $100;
  CG_MSG_CLOSE			         = $101;
  CG_MSG_TIME			           = $102;
  CG_MSG_DATA			           = $110;
  CG_MSG_STREAM_DATA		     = $120;
  CG_MSG_DATAARRAY		       = $150;
  CG_MSG_OBJFAILED		       = $180;
  CG_MSG_PARSEERR			       = $190;
  CG_MSG_TN_BEGIN			       = $200;
  CG_MSG_TN_COMMIT		       = $210;
  CG_MSG_TN_ROLLBACK		     = $220;
  CG_MSG_P2MQ_RANGE_START		 = $1000;
  CG_MSG_P2MQ_TIMEOUT			   = $1 + CG_MSG_P2MQ_RANGE_START;
  CG_MSG_P2REPL_RANGE_START	 = $1100;
  CG_MSG_P2REPL_LIFENUM			 = $10 + CG_MSG_P2REPL_RANGE_START;
  CG_MSG_P2REPL_CLEARDELETED = $11 + CG_MSG_P2REPL_RANGE_START;
  CG_MSG_P2REPL_ONLINE			 = $12 + CG_MSG_P2REPL_RANGE_START;
  CG_MSG_P2REPL_REPLSTATE		 = $15 + CG_MSG_P2REPL_RANGE_START;

  CG_REASON_UNDEFINED     = 0;
  CG_REASON_USER          = 1;
  CG_REASON_ERROR         = 2;
  CG_REASON_DONE          = 3;
  CG_REASON_SNAPSHOT_DONE = 4;

  CG_REPLACT_INSERT = 0;
  CG_REPLACT_UPDATE = 1;
  CG_REPLACT_DELETE = 2;
  CG_REPLACT_INSUPD = 3;

//-- cg_data_lifenum_t flags
  CG_LIFENUM_CHANGE__DONT_CLEAR_STORAGE = $1;
  CG_INDEX_INVALID                      = size_t(-1);
  CG_ID_INVALID                         = uint32_t(-1);
  CG_MAX_REVISON                        = int64_t($7FFFFFFFFFFFFFFF);
  CG_OWNER_UNKNOWN                      = int64_t(0);

//--- Cgate result
type
  CG_RESULT = uint32_t;

//--- General purpose message
  pcg_msg = ^tcg_msg;
  tcg_msg = record
    msg_type: uint32_t;		  // Message type
	  data_size: size_t;	    // Data size
	  data: Pointer;			    // Data pointer
	  owner_id: int64_t;	    // Message owner id
  end;

//--- Time message
  pcg_msg_time = ^tcg_msg_time;
  tcg_msg_time = record
    msg_type: uint32_t;		  // Message type
	  data_size: size_t;	    // Data size
	  data: Pointer;			    // Data pointer
	  owner_id: int64_t;	    // Message owner id
	  usec: uint64_t;		      // Microseconds from epoch
  end;

//--- Stream data message
  pcg_msg_streamdata = ^tcg_msg_streamdata;
  tcg_msg_streamdata = record
    msg_type: uint32_t;			// Message type = CG_MSG_STREAM_DATA
	  data_size: size_t;		  // Data size
	  data: Pointer;				  // Data pointer
	  owner_id: int64_t;	    // Message owner id
    msg_index: size_t;		  // Message number in active scheme
	  msg_id: uint32_t;		    // Unique message ID (if applicable)
	  msg_name: PAnsiChar;	  // Message name in active scheme
    rev: int64_t;			      // Message sequence number
	  num_nulls: size_t;		  // Size of presence map
	  nulls: uint8_tp;        // Presence map. Contains 1 for NULL fields
	  user_id: uint64_t;      // User ID message is intended for
  end;

//--- Data message
//--- union
  u_user_id = record
    user_id: uint32_t;
	  user_ptr: Pointer;
  end;

  pcg_msg_data = ^tcg_msg_data;
  tcg_msg_data = record
    msg_type: uint32_t;			// Message type = CG_MSG_P2REPL_DATA
	  data_size: size_t;		  // Data size
	  data: Pointer;				  // Data pointer
	  owner_id: int64_t;	    // Message owner id
    msg_index: size_t;		  // Message number in active scheme
	  msg_id: uint32_t;		    // Unique message ID (if applicable)
	  msg_name: PAnsiChar;	  // Message name in active scheme
    u_id: u_user_id;        // User message id (the value sent back in reply)
    addr: PAnsiChar;		    // Remote address
	  ref_msg: pcg_msg_data;	// Reference message (not used now)!!!!
  end;

//--- Data array message
  pcg_msg_dataarray = ^tcg_msg_dataarray;
  tcg_msg_dataarray = record
    msg_type: uint32_t;			// Message type = CG_MSG_P2REPL_DATA
	  data_size: size_t;		  // Data size
	  data: Pointer;				  // Data pointer
	  owner_id: int64_t;	    // Message owner id
    msg_index: size_t;		  // Message number in active scheme
	  msg_id: uint32_t;		    // Unique message ID (if applicable)
	  msg_name: PAnsiChar;	  // Message name in active scheme
    u_id: u_user_id;        // User message id (the value sent back in reply)
  	addr: PAnsiChar;		    // Remote address
	  ref_msg: pcg_msg_data;  // Reference message (not used now) !!!
	  cnt: uint32_t;          // Count
  end;

//--- User message callback function
  cg_LISTENER_CB_t = function (hConn, hListener: THandle; msg: pcg_msg; data: pointer): CG_RESULT; stdcall;
//typedef CG_RESULT (CG_API *CG_LISTENER_CB)(cg_conn_t* conn, cg_listener_t* listener, struct cg_msg_t* msg, void* data);

//--- List of key-value pairs
  pcg_value_pair = ^tcg_value_pair;
  tcg_value_pair = record
    next: pcg_value_pair;   // Pointer to the next list entry
    key: PAnsiChar;         // Key, required
    value: PAnsiChar;       // Value, may be null
  end;

//--- Field value description
  pcg_field_value_desc = ^tcg_field_value_desc;
  tcg_field_value_desc = record
    next: pcg_field_value_desc; // pointer to the next value
    name: PAnsiChar;        // name
    desc: PAnsiChar;        // Description. May be NULL if not defined
    value: Pointer;         // pointer to value
    mask: Pointer;          // used for integer fields only (i[1-8], u[1-8]) a mask that defines range of bits used by the value
  end;

//--- Message description*
  pcg_message_desc = ^tcg_message_desc;

//--- Message field description
  pcg_field_desc = ^tcg_field_desc;
  tcg_field_desc = record
    next: pcg_field_desc;   // pointer to the next value
    id: uint32_t;           // Field ID. May be 0 if not defined
    name: PAnsiChar;        // Field name
    desc: PAnsiChar;        // Field description May be NULL if not defined
    field_type: PAnsiChar;  // field type
    size: size_t;           // size in bytes
    offset: size_t;         // offset from buffer beginning
    def_value: Pointer;     // Pointer to default value of the field Points to the buffer of size "size" May be NULL if no default value is defined
    num_values: size_t;     // Number of values for the field values
    values: pcg_field_value_desc; // Pointer to the list of values which can be taken by the field
    hints: pcg_value_pair;  // Field options
    max_count: size_t;      // Maximum number of fields, 1 by default
    count_field: pcg_field_desc; // Link to description of count field
    type_msg: pcg_message_desc;  // Pointer to message description for type = 'm' fields
  end;

//--- Index component description
  pcg_indexfield_desc = ^tcg_indexfield_desc;
  tcg_indexfield_desc = record
    next: pcg_indexfield_desc; // Pointer to the next value
    field: pcg_field_desc;  // points to field of index
    sort_order: uint32_t;   // sort order (0 for asc, 1 for desc)
  end;

//--- Index description
  pcg_index_desc = ^tcg_index_desc;
  tcg_index_desc = record
    next: pcg_index_desc;   // Pointer to the next index
    num_fields: size_t;     // Number of fields in the index
    fields: pcg_indexfield_desc; // Pointer to the first index component
    name: PAnsiChar;        // name
    desc: PAnsiChar;        // desc may be NULL if not defined
    hints: pcg_value_pair;  // index hints
  end;


//---* Message description
  tcg_message_desc = record
    next: pcg_message_desc; // Pointer to the next message
    size: size_t;           // Message data block size
    num_fields: size_t;     // Number of fields in the message
    fields: pcg_field_desc; // Pointer to the first message description
    id: uint32_t;           // Message unique ID May be 0 if not defined
    name: PAnsiChar;        // Message name May be NULL if not defined
    desc: pAnsiChar;        // Description May be NULL if not defined
    hints: pcg_value_pair;  // Message hints May be NULL if not defined
    num_indices: size_t;    // Number of indices for the message
    indices: pcg_index_desc; // Pointer to the first index description
    align: size_t;          // Size of alignment
  end;

//--- Scheme description
  pcg_scheme_desc = ^tcg_scheme_desc;
  tcg_scheme_desc = record
    scheme_type: uint32_t;  // Scheme type 1 for binary scheme which is the only type supported now
    features: uint32_t;     // Scheme features (combination of CG_SCHEME_BIN_* constants)
    num_messages: size_t;   // Number of messages in the scheme
    messages: pcg_message_desc; // Pointer to the first message description
    hints: pcg_value_pair;  // Scheme options
  end;

//--- Date-time type
  pcg_time = ^tcg_time;
  tcg_time = record
    year: uint16_t;         // Year
	  month: uint8_t;         // Month of year (1-12)
	  day: uint8_t;           // Day of month (1-31)
	  hour: uint8_t;          // Hour (0-23)
	  minute: uint8_t;        // Minute (0-59)
	  second: uint8_t;        // Second (0-59)
	  msec: uint16_t;         // Millisecond (0-999)
  end;

//--- Structure that describes data format  for CG_MSG_P2REPL_CLEARDELETED message
  pcg_data_cleardeleted = ^tcg_data_cleardeleted;
  tcg_data_cleardeleted = record
    table_idx: uint32_t;
	  table_rev: int64_t;
  end;

//--- Repl act
  pcg_repl_act = ^tcg_repl_act;
  tcg_repl_act = record
    act: uint8_t;           // INSERT, UPDATE, DELETE
	  idx_id: uint8_t;
	  reserved: array[0..5] of AnsiChar;
  end;

//--- Lifenum
  pcg_data_lifenum = ^tcg_data_lifenum;
  tcg_data_lifenum = record
    life_number: uint32_t;
	  flags: uint32_t;        // CG_LIFENUM_CHANGE_
  end;

//--- Zblock
  pcg_zblock = ^tcg_zblock;
  tcg_zblock = record
    size: uint32_t;
	  data_ptr: uint64_t;
  end;
// -----------------------------------------------------------------------------
// environment
// -----------------------------------------------------------------------------
//--- Initialize environment
  cg_env_open_t = function(const settings: PAnsiChar): CG_RESULT; stdcall;

//--- Deinitialize environment
  cg_env_close_t = function(): CG_RESULT; stdcall;
{
 Create new connection
 @param settings Connection initialization string
 @param connptr Where pointer to new connection will be stored

 Connection initialization string has the following format:
 "TYPE://HOST:PORT&param1=value1&param2=value2..."
 ,where
 - TYPE defines connection type
   - "p2tcp" means TCP connection with Plaza-2 router process
   - "p2lrcpq" meanes shared memory connection with Plaza-2 router process
 - HOST is host of the machine where Plaza-2 router runs
 - PORT is port of Plaza-2 router
 - param1 ... paramN - parameters which depend on type of connection.

 Parameters for "p2tcp" and "p2lrpcq" connections:
 - app_name - a string, Plaza-2 application name for connection
            each connection must have unique name
 - timeout - a number, timeout of Plaza-2 router connection, ms
 - local_timeout - a number, timeout of Plaza-2 router interactions, ms
 }
//--- New connection
  cg_conn_new_t = function(settings: PAnsiChar; var hConn: THandle): CG_RESULT; stdcall;

//--- Destroys connection
  cg_conn_destroy_t = function (hConn: THandle): CG_RESULT; stdcall;

{
  Open connection
  @param settings Connection open string
  Parameter "settigns" is not used for both p2tcp and p2lrpcq connection types.
}
  cg_conn_open_t = function(hConn: THandle; const settings: PAnsiChar): CG_RESULT; stdcall;

//--- Close connection
  cg_conn_close_t = function(hConn: THandle): CG_RESULT; stdcall;

//--- Get connection state
  cg_conn_getstate_t = function(hConn: THandle; var state: uint32_t): CG_RESULT; stdcall;

//--- Process one iteration of connection internal logic
  cg_conn_process_t = function(hConn: THandle; timeout: uint32_t; reserved: Pointer): CG_RESULT; stdcall;


{ Create new listener
   @param conn Connection to create listener on
   @param settings Connection initialization string
   @param cb Pointer to user message callback
   @param data User data that will be passed as one of the callback parameters
   @param lsnptr Where pointer to new listener will be stored

   There are following types of listeners supported:
   p2repl - datastream replication client
   p2mqreply - messages reply listener
   p2ordbook - optimized orderbook datastream listener

   Initialization string for "p2repl":
   p2repl://STREAM[;scheme=SCHEMEURL]
     * STREAM - datastream name
	 * SCHEMEURL - scheme URL in format "|SRC|PATH|NAME"
	   where SRC is either FILE or MQ,
	         PATH is path to scheme INI file for FILE source
			 PATH is scheme service name for MQ soruce
			 NAME is name of the scheme in file or scheme service

   Initialization string for "p2mqreply":
   p2mqreply://;ref=REFERENCE
     * REFERENCE - name of the publisher that is used to send messages

   Initialization string for "p2ordbook":
   p2ordbook://STREAM;snapshot=STREAMSS[;scheme=SCHEME][;snapshot.scheme=SCHEMESS]
     * STREAM - name of orderbook online stream
	 * STREAMSS - name of orderbook snapshot stream
	 * SCHEME - scheme URL for online stream
	 * SCHEMESS - scheme URL for snapshot stream

 }
//--- New listener
  cg_lsn_new_t = function(hConn: THandle; const settings: PAnsiChar; cb: cg_LISTENER_CB_t; data: Pointer; var hLsn: THandle): CG_RESULT; stdcall;

//--- Destroy listener
  cg_lsn_destroy_t = function(hLsn: THandle): CG_RESULT; stdcall;

//--- Get listener state
  cg_lsn_getstate_t = function(hLsn: THandle; var state: uint32_t): CG_RESULT; stdcall;

{	@param settings Listener open string

	Open string for "p2repl":
	[mode=STREAMTYPES][;replstate=REPLSTATE]
		* STREAMTYPES - stream mode ('snapshot', 'online', 'snapshot+online')
		* REPLSTATE - previously saved stream stated (string received in CG_MSG_P2REPL_REPLSTATE message)

	Parameter "settings" is ignored for other types of listeners.
}
//--- Open listener
  cg_lsn_open_t = function(hLsn: THandle; const settings: PAnsiChar): CG_RESULT; stdcall;

//-- Close listener
  cg_lsn_close_t = function(hLsn: THandle): CG_RESULT; stdcall;

//--- Get current data scheme for listener
  cg_lsn_getscheme_t = function(hLsn: THandle; var desc: pcg_scheme_desc): CG_RESULT; stdcall;

{	There are following types of publishers supported::
	p2mq - posts messages to specified Plaza-2 service

	Initialization string for "p2mq":
	p2mq://SERVICE[;scheme=SCHEMEURL][;timeout=TIMEOUT][;name=NAME]
		* SERVICE - destination service name
		* SCHEMEURL - scheme URL
		* TIMEOUT - reply timeout, ms
		* NAME - unique publisher name to be referenced by p2mqreply listener
}
//--- Create new publisher
  cg_pub_new_t = function(hConn: THandle; const settings: PAnsiChar; var hPub: THandle): CG_RESULT; stdcall;

//--- Destroy publisher
  cg_pub_destroy_t = function(hPub: THandle): CG_RESULT; stdcall;

{  @param settings Publisher open string
  Parameter "settings" is not used currently and should be empty.
}
//--- Open publisher
  cg_pub_open_t = function(hPub: THandle; const settings: PAnsiChar): CG_RESULT; stdcall;

//--- Close publisher
  cg_pub_close_t = function(hPub: THandle): CG_RESULT; stdcall;

//--- Get publisher state
  cg_pub_getstate_t = function(hPub: THandle; var state: uint32_t): CG_RESULT; stdcall;

//--- Get data scheme description
  cg_pub_getscheme_t = function(hPub: THandle; var desc: pcg_scheme_desc): CG_RESULT; stdcall;


{/** Create new message for publishing
   @param id_type CG_KEY_INDEX Create message using its index in current publisher scheme (id points to uint32_t)
   @param id_type CG_KEY_ID Create message using its unique ID in publisher scheme (id points to uint32_t)
   @param id_type CG_KEY_NAME Create message using its name in current publisher scheme (id points to 0-terminated string)
}
//--- Create new message for publishing
  cg_pub_msgnew_t = function(hPub: THandle; id_type: uint32_t; const id: Pointer; var msgptr: pcg_msg): CG_RESULT; stdcall;

//--- Free message
  cg_pub_msgfree_t = function(hPub: THandle; msgptr: pcg_msg): CG_RESULT; stdcall;

{/** Post message
   @param flags Flags of post operation
          CG_PUB_NEEDREPLY - if message reply is expected
}
//--- Post message
  cg_pub_post_t = function(hPub: THandle; msgptr: pcg_msg; flags: uint32_t): CG_RESULT; stdcall;

{/**
 Return string representation for error code.
 @param errCode CGate error code
}
  cg_err_getstr_t = function(errCode: CG_RESULT): PAnsiChar; stdcall;

{ //TODO:???????????????????????????
 Log user message
 @param fmt Format string (printf-like)
 @param ... Parameters

CG_EXPORT CG_RESULT CG_API cg_log_trace(const char* fmt, ...);
CG_EXPORT CG_RESULT CG_API cg_log_debug(const char* fmt, ...);
CG_EXPORT CG_RESULT CG_API cg_log_info(const char* fmt, ...);
CG_EXPORT CG_RESULT CG_API cg_log_error(const char* fmt, ...);
}
{/**
 Log user message
 @param fmt Format string (printf-like)
 @param ... Parameters
 }
  cg_log_va_trace_t = function(const fmt: PAnsiChar; va: va_list): CG_RESULT; stdcall;
  cg_log_va_debug_t = function(const fmt: PAnsiChar; va: va_list): CG_RESULT; stdcall;
  cg_log_va_info_t = function(const fmt: PAnsiChar; va: va_list): CG_RESULT; stdcall;
  cg_log_va_error_t = function(const fmt: PAnsiChar; va: va_list): CG_RESULT; stdcall;

{/**
 Log user message
 For languages that do not support variable number of arguments
 @param str String to log
 */}
  cg_log_tracestr_t = function(const logstr: PAnsiChar): CG_RESULT; stdcall;
  cg_log_debugstr_t = function(const logstr: PAnsiChar): CG_RESULT; stdcall;
  cg_log_infostr_t = function(const logstr: PAnsiChar): CG_RESULT; stdcall;
  cg_log_errorstr_t = function(const logstr: PAnsiChar): CG_RESULT; stdcall;

{/**
 Dump specified message as a text. Message data will be parsed if data scheme is specified as paramater.
 This function is designed for debugging purposes.
 @param msg Message to dump
 @param scheme Data scheme (may be NULL)
 @param buffer Pointer to preallocated buffer where to store dump
 @param bufsize Points to variable holding buffer size.

 If the buffer is too small for message dump, CG_ERR_BUFFERTOOSMALL error code will be returned and
 bufsize will be filled with desired buffer size.
 */ }
  cg_msg_dump_t = function(msgptr: Pointer; scheme: pcg_scheme_desc; buffer: pAnsiChar; var bufsize: size_t): CG_RESULT; stdcall;

{/**
 Return BCD value as integer part and position of decimal point.
 Returned values can be used the following way:
 val = bigdecimal(intpart) / 10^scale

 @param bcd Point to BCD value
 @param intpart Where to store integer part
 @param scale Where to store scale
 */}
  cg_bcd_get_t = function(bcd: Pointer; var intpart: int64_t; var scale: int8_t): CG_RESULT; stdcall;

{/**
 Return data represented as text
 @param type Field type (the same format as used in schemes)
 @param data Data pointer
 @param buffer Pointer to buffer where string will be stored
 @param bufsize Points to variable holding buffer size

 If the buffer is too small for string representation, CG_ERR_BUFFERTOOSMALL error code will be
 returned and bufsize will be filled with desired buffer size.
 */}
  cg_getstr_t = function(const val_type: PAnsiChar; const data: Pointer; var buffer: PAnsiChar; var bufsize: size_t): CG_RESULT; stdcall;

{/**
 Return 0 on success
 @param name name of the component
 @param major pointer to int where function put major version of the component
 @param minor pointer to int where function put minor version of the component
 @param patch pointer to int where function put patch version of the component
 */}
  cg_env_getcomp_ver_t = function( const name: PAnsiChar; var major: integer; var minor: integer; var patch: integer ): CG_RESULT; stdcall;

{/**
User exit event handler
@param data user context
 */ }
  CG_EXITEVENT_CB = procedure(data: Pointer); stdcall;

{/**
Register User exit event handler
@param cb user exit event handler
@param cd_data user context
 */}
  cg_env_register_exithandler_t = function(cb: CG_EXITEVENT_CB; cb_data: Pointer ):CG_RESULT; stdcall;

var
  cg_env_open: cg_env_open_t;
  cg_env_close: cg_env_close_t;
  cg_conn_new: cg_conn_new_t;
  cg_conn_destroy: cg_conn_destroy_t;
  cg_conn_open: cg_conn_open_t;
  cg_conn_close: cg_conn_close_t;
  cg_conn_getstate: cg_conn_getstate_t;
  cg_conn_process: cg_conn_process_t;
  cg_lsn_new: cg_lsn_new_t;
  cg_lsn_destroy: cg_lsn_destroy_t;
  cg_lsn_getstate: cg_lsn_getstate_t;
  cg_lsn_open: cg_lsn_open_t;
  cg_lsn_close: cg_lsn_close_t;
  cg_lsn_getscheme: cg_lsn_getscheme_t;
  cg_pub_new: cg_pub_new_t;
  cg_pub_destroy: cg_pub_destroy_t;
  cg_pub_open: cg_pub_open_t;
  cg_pub_close: cg_pub_close_t;
  cg_pub_getstate: cg_pub_getstate_t;
  cg_pub_getscheme: cg_pub_getscheme_t;
  cg_pub_msgnew: cg_pub_msgnew_t;
  cg_pub_msgfree: cg_pub_msgfree_t;
  cg_pub_post: cg_pub_post_t;
  cg_err_getstr: cg_err_getstr_t;
  cg_msg_dump: cg_msg_dump_t;
  cg_bcd_get: cg_bcd_get_t;
  cg_getstr: cg_getstr_t;
  cg_log_tracestr: cg_log_tracestr_t;
  cg_log_debugstr: cg_log_debugstr_t;
  cg_log_infostr: cg_log_infostr_t;
  cg_log_errorstr: cg_log_errorstr_t;
  cg_log_va_trace: cg_log_va_trace_t;
  cg_log_va_debug: cg_log_va_debug_t;
  cg_log_va_info: cg_log_va_info_t;
  cg_log_va_error: cg_log_va_error_t;
  cg_env_getcomp_ver: cg_env_getcomp_ver_t;
  cg_env_register_exithandler: cg_env_register_exithandler_t;
  cg_LISTENER_CB: cg_LISTENER_CB_t;


implementation

end.
